"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.afm` subpackage.

Routine listings
----------------
get(key=None)
    Get the value of one or more configuration options.
set(dictionary={}, \*\*kwargs)
    Set the value of one or more configuration options.

See also
--------
magni.utils.config : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by `magni.utils.config`
and assigns handles for the `get` and `set` methods of that class instance. The
configuration options are the following:

algorithm : {'iht', 'sl0'}
    The compressed sensing reconstruction algorithm to use (the default is
    'iht').

"""

from __future__ import division

from magni.utils.config import Configger as _Configger


_configger = _Configger({'algorithm': 'iht'},
                        {'algorithm': {'type': str,
                                       'value_in': ['iht', 'sl0']}})

get = _configger.get
set = _configger.set



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing atomic force miscroscopy specific functionality.

The present subpackage includes functionality for handling AFM files and data
and functionality for utilizing the other subpackages for such AFM data.

Routine listings
----------------
config
    Module providing configuration options for this subpackage.
io
    Module providing input/output functionality for MI files.
reconstruction
    Module providing reconstruction and analysis of reconstructed images.
types
    Module providing data container classes for MI files.

"""

from magni.afm import config
from magni.afm import io
from magni.afm import reconstruction
from magni.afm import types



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the original SL0 reconstruction algorithm.

Routine listings
----------------
run(y, A)
    Run the original SL0 reconstruction algorithm.

See Also
--------
magni.cs.reconstruction.sl0.config : Configuration options.

Notes
-----
| The original SL0 reconstruction algorithm is described in [1]_ and [2]_.
|     For delta < 0.55: Standard projection algorithm by Mohimani et. al [1]_
|     For delta >= 0.55: Standard constraint elimination algorithm by Cui et.
      al. [2]_

References
----------
.. [1] H. Mohimani, M. Babaie-Zadeh, and C. Jutten, "A Fast Approach for
   Overcomplete Sparse Decomposition Based on Smoothed l0 Norm", *IEEE
   Transactions on Signal Processing*, vol. 57, no. 1, pp. 289-301, Jan. 2009.
.. [2] Z. Cui, H. Zhang, and W. Lu, "An Improved Smoothed l0-norm Algorithm
   Based on Multiparameter Approximation Function", *in 12th IEEE International
   Conference on Communication Technology (ICCT)*, Nanjing, China, Nov. 11-14,
   2011, pp. 942-945.

"""

from __future__ import division

import numpy as np
import scipy

from magni.cs.reconstruction.sl0 import config as _config


def run(y, A):
    """
    Run the SL0 reconstruction algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    See Also
    --------
    _run_proj : The original projection algorithm.
    _run_feas : The original constraint elimination algorithm.

    Examples
    --------
    For example, recovering a vector from random measurements

    >>> from magni.cs.reconstruction.sl0._original import run
    >>> np.random.seed(seed=6021)
    >>> A = 1 / np.sqrt(80) * np.random.randn(80, 200)
    >>> x = np.zeros((200, 1))
    >>> x[:10] = 1
    >>> y = A.dot(x)
    >>> x_hat = run(y, A)
    >>> x_hat[:12]
    array([[  9.99840757e-01],
           [  9.99849856e-01],
           [  9.99955438e-01],
           [  9.99966334e-01],
           [  1.00010956e+00],
           [  1.00000432e+00],
           [  9.99995701e-01],
           [  1.00016335e+00],
           [  9.99927317e-01],
           [  9.99841626e-01],
           [ -3.01131370e-05],
           [  4.10127956e-06]])
    >>> (np.abs(x_hat) > 1e-2).sum()
    10

    """

    if A.shape[0] / A.shape[1] < 0.55:
        x = _run_proj(y, A)
    else:
        x = _run_feas(y, A)

    return x


def _run_feas(y, A):
    """
    Run the original *feasibility* SL0 reconstruction algorithm.

    This function implements the algorithm with a search on the feasible set.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    """

    param = _config.get()
    sigma_min = param['sigma_min']
    L = int(np.round(param['L']))
    mu = param['mu']
    sigma_update = param['sigma_update']

    Q, R = scipy.linalg.qr(A.T, mode='economic')
    IP = np.eye(A.shape[1]) - Q.dot(Q.T)
    x = Q.dot(scipy.linalg.solve_triangular(R, y, trans='T'))
    sigma = param['precision_float'](2) * np.abs(x).max()

    while sigma > sigma_min:
        for j in range(L):
            d = np.exp(-x ** 2 / (2 * sigma ** 2)) * x
            x = x - mu * IP.dot(d)  # Search on feasible set

        sigma = sigma * sigma_update

    return x


def _run_proj(y, A):
    """
    Run the original *projection* SL0 reconstruction algorithm.

    This function implements the algorithm with an unconstrained gradient step
    followed by a projection back onto the feasible set.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    """

    param = _config.get()
    sigma_min = param['sigma_min']
    L = int(np.round(param['L']))
    mu = param['mu']
    sigma_update = param['sigma_update']

    Q, R = scipy.linalg.qr(A.T, mode='economic')
    A_pinv = Q.dot(scipy.linalg.inv(R.T))
    x = A_pinv.dot(y)
    sigma = param['precision_float'](2) * np.abs(x).max()

    while sigma > sigma_min:
        for j in range(L):
            d = x * np.exp(-x ** 2 / (2 * sigma ** 2))
            x = x - mu * d
            x = x - A_pinv.dot(A.dot(x) - y)  # Projection

        sigma = sigma * sigma_update

    return x



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for evaluation of image reconstruction quality.

Routine listings
----------------
calculate_mse(x_org, x_recons)
    Function to calcualte Mean Squared Error (MSE).
calculate_psnr(x_org, x_recons, peak)
    Function to calculate Peak Signal to Noise Ratio (PSNR).
calculate_retained_energy(x_org, x_recons)
    Function to calculate the percentage of energy retained in reconstruction.

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate
from magni.utils.validation import validate_ndarray as _validate_ndarray


@_decorate_validation
def _validate_calculate_mse(x_org, x_recons):
    """
    Validate the `calculate_mse` function.

    See Also
    --------
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(x_org, 'x_org', {})
    _validate_ndarray(x_recons, 'x_recons', {'shape': x_org.shape})


def calculate_mse(x_org, x_recons):
    r"""
    Calculate Mean Squared Error (MSE) between `x_recons` and `x_org`.

    Parameters
    ----------
    x_org : ndarray
        Array of original values.
    x_recons : ndarray
        Array of reconstruction values.

    Returns
    -------
    mse : float
        Mean Squared Error (MSE).

    Notes
    -----
    The Mean Squared Error (MSE) is calculated as:

    .. math::

         \frac{1}{N} \cdot \sum(x_{org} - x_{recons})^2

    where `N` is the number of entries in `x_org`.

    Examples
    --------
    For example,

    >>> from magni.imaging.evaluation import calculate_mse
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_recons = np.ones((2,2))
    >>> calculate_mse(x_org, x_recons)
    1.5

    """

    _validate_calculate_retained_energy(x_org, x_recons)

    return ((x_org - x_recons) ** 2).mean()


@_decorate_validation
def _validate_calculate_psnr(x_org, x_recons, peak):
    """
    Validate the `calculate_psnr` function.

    See Also
    --------
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(x_org, 'x_org', {})
    _validate_ndarray(x_recons, 'x_recons', {'shape': x_org.shape})
    _validate(peak, 'peak', {'type_in': [float, int], 'min': 0})


def calculate_psnr(x_org, x_recons, peak):
    r"""
    Calculate Peak Signal to Noise Ratio (PSNR) between `x_recons` and `x_org`.

    Parameters
    ----------
    x_org : ndarray
        Array of original values.
    x_recons : ndarray
        Array of reconstruction values.
    peak : int or float
        Peak value.

    Returns
    -------
    psnr : float
        Peak Signal to Noise Ratio (PSNR) in dB.

    Notes
    -----
    The PSNR is as calculated as

    .. math::

         10 \cdot \log_{10}\left(\frac{peak^2}{ 1/N \cdot \sum(x_{org} -
         x_{recons})^2}\right)

    where `N` is the number of entries in `x_org`.

    Examples
    --------
    For example,

    >>> from magni.imaging.evaluation import calculate_psnr
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_recons = np.ones((2,2))
    >>> peak = 3
    >>> calculate_psnr(x_org, x_recons, peak)
    7.7815125038364368

    """

    _validate_calculate_psnr(x_org, x_recons, peak)

    return 10 * np.log10(peak**2 / ((x_org - x_recons) ** 2).mean())


@_decorate_validation
def _validate_calculate_retained_energy(x_org, x_recons):
    """
    Validate the `calculate_retained_energy` function.

    See Also
    --------
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(x_org, 'x_org', {})
    _validate_ndarray(x_recons, 'x_recons', {'shape': x_org.shape})


def calculate_retained_energy(x_org, x_recons):
    r"""
    Calculate percentage of energy retained in reconstruction.

    Parameters
    ----------
    x_org : ndarray
        Array of original values.
    x_recons : ndarray
        Array of reconstruction values.

    Returns
    -------
    energy : float
        Percentage of retained energy in reconstruction.

    Notes
    -----
    The retained energy is as calculated as

    .. math::
         \frac{\sum x_{recons}^2}{\sum x_{org}^2} \cdot 100\%

    Examples
    --------
    For example,

    >>> from magni.imaging.evaluation import calculate_retained_energy
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_recons = np.ones((2,2))
    >>> calculate_retained_energy(x_org, x_recons)
    28.571428571428569

    """

    _validate_calculate_retained_energy(x_org, x_recons)

    return (x_recons ** 2).sum() / (x_org ** 2).sum() * 100



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the modified SL0 reconstruction algorithm.

Routine listings
----------------
run(y, A)
    Run the modified SL0 reconstruction algorithm.

See Also
--------
magni.cs.reconstruction.sl0.config : Configuration options.

Notes
-----
| The modified SL0 reconstruction algorithm is described in [1]_.
|     For delta < 0.55: Modified projection algorithm
|     For delta >= 0.55: Modified constraint elimination algorithm

References
----------
.. [1] C. S. Oxvig, P. S. Pedersen, T. Arildsen, and T. Larsen, "Surpassing the
   Theoretical 1-norm Phase Transition in Compressive Sensing by Tuning the
   Smoothed l0 Algorithm", *in IEEE International Conference on Acoustics,
   Speech and Signal Processing (ICASSP)*, Vancouver, Canada, May 26-31, 2013,
   pp. 6019-6023.

"""

from __future__ import division

import numpy as np
import scipy

from magni.cs.reconstruction.sl0 import config as _config


def run(y, A):
    """
    Run the modified SL0 reconstruction algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    See Also
    --------
    _run_proj : The original projection algorithm.
    _run_feas : The original constraint elimination algorithm.

    Examples
    --------
    For example, recovering a vector from random measurements

    >>> from magni.cs.reconstruction.sl0._modified import run
    >>> np.random.seed(seed=6021)
    >>> A = 1 / np.sqrt(80) * np.random.randn(80, 200)
    >>> x = np.zeros((200, 1))
    >>> x[:10] = 1
    >>> y = A.dot(x)
    >>> x_hat = run(y, A)
    >>> x_hat[:12]
    array([[  9.99997941e-01],
           [  9.99999463e-01],
           [  1.00000090e+00],
           [  9.99998622e-01],
           [  1.00000078e+00],
           [  9.99998433e-01],
           [  1.00000025e+00],
           [  1.00000346e+00],
           [  1.00000088e+00],
           [  9.99995474e-01],
           [ -4.12075673e-07],
           [  2.17244596e-07]])
    >>> (np.abs(x_hat) > 1e-2).sum()
    10

    """

    if A.shape[0] / A.shape[1] < 0.55:
        x = _run_proj(y, A)
    else:
        x = _run_feas(y, A)

    return x


def _calc_sigma_start(delta):
    """
    Calculate the initial sigma factor for a given indeterminacy.

    Parameters
    ----------
    delta : float
        The indeterminacy, m / n, of a system of equations of size m x n.

    Returns
    -------
    sigma_start : float
        The initial sigma factor for the given indeterminacy.

    """

    return 1.0 / (2.75 * delta)


def _run_feas(y, A):
    """
    Run the modified *feasibility* SL0 reconstruction algorithm.

    This function implements the algorithm with a search on the feasible set.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    """

    param = _config.get()

    sigma_update = param['sigma_update']
    sigma_min = param['sigma_min']

    L = param['L']
    L_update = param['L_update']

    mu_start = param['mu_start']
    mu_end = param['mu_end']

    epsilon = param['epsilon']

    Q, R = scipy.linalg.qr(A.T)
    Q1 = Q[:, :A.shape[0]]
    Q2 = Q[:, A.shape[0]:]
    R = R[:R.shape[1], :]
    x = Q1.dot(scipy.linalg.solve_triangular(R, y, trans='T'))

    mult = _calc_sigma_start(A.shape[0] / A.shape[1])
    sigma = mult * np.abs(x).max()

    x_zeros = np.zeros(x.shape)
    i = 0

    while sigma > sigma_min:
        if i < 4 or mult * sigma_update**i > 0.75:
            if A.shape[0] / A.shape[1] <= 0.5:
                mu = 50 * mu_start
            else:
                mu = mu_start
        else:
            mu = mu_end

        x_prev = x_zeros
        j = 0

        while scipy.linalg.norm(x - x_prev) > sigma * epsilon and j <= L:
            x_prev = x.copy()

            d = np.exp(-(x ** 2) / (2 * sigma ** 2)) * x
            nabla = Q2.T.dot(d)
            x = x - Q2.dot(mu * nabla)  # Search on feasible set

            j = j + 1

        sigma = sigma * sigma_update
        L = L * L_update
        i = i + 1

    return x


def _run_proj(y, A):
    """
    Run the original *projection* SL0 reconstruction algorithm.

    This function implements the algorithm with an unconstrained gradient step
    followed by a projection back onto the feasible set.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    """

    param = _config.get()

    sigma_update = param['sigma_update']
    sigma_min = param['sigma_min']

    L = param['L']
    L_update = param['L_update']

    mu_start = param['mu_start']
    mu_end = param['mu_end']

    epsilon = param['epsilon']

    Q, R = scipy.linalg.qr(A.T, mode='economic')
    A_pinv = Q.dot(scipy.linalg.inv(R.T))
    x = A_pinv.dot(y)

    mult = _calc_sigma_start(A.shape[0] / A.shape[1])
    sigma = mult * np.abs(x).max()

    x_zeros = np.zeros(x.shape)
    i = 0

    while sigma > sigma_min:
        if i < 4 or mult * sigma_update**i > 0.75:
            if A.shape[0] / A.shape[1] <= 0.5:
                mu = 50 * mu_start
            else:
                mu = mu_start
        else:
            mu = mu_end

        x_prev = x_zeros
        j = 0

        while scipy.linalg.norm(x - x_prev) > sigma * epsilon and j <= L:
            x_prev = x.copy()

            d = np.exp(-(x ** 2) / (2 * sigma ** 2)) * x
            x = x - mu * d
            x = x - A_pinv.dot(A.dot(x) - y)  # Projection

            j = j + 1

        sigma = sigma * sigma_update
        L = L * L_update
        i = i + 1

    return x



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing AFM image reconstruction and analysis of reconstructed images.

Routine listings
----------------
analyse(x, Phi, Psi)
    Sample an image, reconstruct it, and analyse the reconstructed image.
reconstruct(y, Phi, Psi)
    Reconstruct an image from compressively sensed measurements.

"""

from __future__ import division

import numpy as np

from magni.afm import config as _config
from magni.cs.reconstruction import iht as _iht
from magni.cs.reconstruction import sl0 as _sl0
from magni.imaging import evaluation as _eval
from magni.utils.matrices import Matrix as _Matrix
from magni.utils.matrices import MatrixCollection as _MatrixC
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_ndarray as _validate_ndarray


@_decorate_validation
def _validate_analyse(x, Phi, Psi):
    """
    Validate the `analyse` function.

    See Also
    --------
    analyse : The validated function.
    magni.utils.validation.validate_ndarray : Validation.
    magni.utils.validation.validate_matrix : Validation.

    """

    try:
        _validate_ndarray(Phi, 'Phi')
    except TypeError:
        if not isinstance(Phi, _Matrix) and not isinstance(Phi, _MatrixC):
            raise TypeError('Phi must be a matrix.')

    try:
        _validate_ndarray(Psi, 'Psi')
    except TypeError:
        if not isinstance(Psi, _Matrix) and not isinstance(Psi, _MatrixC):
            raise TypeError('Psi must be a matrix.')

    if Phi.shape[1] != Psi.shape[0]:
        raise ValueError('Phi and Psi must have compatible shapes.')

    _validate_ndarray(x, 'x', {'shape': (Phi.shape[1], 1)})


def analyse(x, Phi, Psi):
    """
    Sample an image, reconstruct it, and analyse the reconstructed image.

    Parameters
    ----------
    x : numpy.ndarray
        The original image vector.
    Phi : magni.utils.matrices.Matrix or numpy.matrix
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix or numpy.matrix
        The dictionary.

    Returns
    -------
    x : numpy.ndarray
        The reconstructed image vector.

    See Also
    --------
    magni.afm.config : Configuration options.
    magni.imaging.evaluation : Image reconstruction quality evaluation.

    Examples
    --------
    Prior to the actual example, data is loaded and a measurement matrix and a
    dictionary are defined. First, the example MI file provided with the
    package is loaded:

    >>> from magni.afm.reconstruction import analyse
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> mi_file = magni.afm.io.read_mi_file(path)
    >>> mi_buffer = mi_file.get_buffer('Topography')[0]
    >>> mi_data = mi_buffer.get_data()
    >>> x = magni.imaging.mat2vec(mi_data)

    Next, a measurement matrix is defined. This matrix is equal to the matrix
    generated by running `np.eye(len(x))[::2, :]` but for speed, the matrix is
    instead defined with fast operations:

    >>> def Phi_A(x):
    ...     y = x[::2]
    ...     return y
    >>> def Phi_T(y):
    ...     x = np.zeros((2 * len(y), 1))
    ...     x[::2] = y
    ...     return x
    >>> Phi = magni.utils.matrices.Matrix(Phi_A, Phi_T, (),
    ...                                   (int(len(x) / 2), len(x)))

    Next, a dictionary is defined. This dictionary is the DCT basis likewise
    defined with fast operations:

    >>> Psi = magni.imaging.dictionaries.get_DCT(mi_data.shape)

    Finally, the actual example:

    >>> print('MSE: {:.2f}, PSNR: {:.2f}'.format(*analyse(x, Phi, Psi)))
    MSE: 0.23, PSNR: 5.92

    """

    _validate_analyse(x, Phi, Psi)

    y = Phi.dot(x)
    x_hat = reconstruct(y, Phi, Psi)

    mse = _eval.calculate_mse(x, x_hat)
    psnr = _eval.calculate_psnr(x, x_hat, float(np.abs(x).max()))

    return (mse, psnr)


@_decorate_validation
def _validate_reconstruct(y, Phi, Psi):
    """
    Validate the `reconstruct` function.

    See Also
    --------
    reconstruct : The validated function.
    magni.utils.validation.validate_ndarray : Validation.
    magni.utils.validation.validate_matrix : Validation.

    """

    try:
        _validate_ndarray(Phi, 'Phi')
    except TypeError:
        if not isinstance(Phi, _Matrix) and not isinstance(Phi, _MatrixC):
            raise TypeError('Phi must be a matrix.')

    try:
        _validate_ndarray(Psi, 'Psi')
    except TypeError:
        if not isinstance(Psi, _Matrix) and not isinstance(Psi, _MatrixC):
            raise TypeError('Psi must be a matrix.')

    if Phi.shape[1] != Psi.shape[0]:
        raise ValueError('Phi and Psi must have compatible shapes.')

    _validate_ndarray(y, 'y', {'shape': (Phi.shape[0], 1)})


def reconstruct(y, Phi, Psi):
    """
    Reconstruct an image from compressively sensed measurements.

    Parameters
    ----------
    y : numpy.ndarray
        The measurement vector.
    Phi : magni.utils.matrices.Matrix or numpy.matrix
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix or numpy.matrix
        The dictionary.

    Returns
    -------
    x : numpy.ndarray
        The reconstructed image vector.

    See Also
    --------
    magni.afm.config : Configuration options.
    magni.cs.reconstruction : Compressed sensing reconstruction algorithms.

    Examples
    --------
    Prior to the actual example, data is loaded and a measurement matrix and a
    dictionary are defined. First, the example MI file provided with the
    package is loaded:

    >>> from magni.afm.reconstruction import reconstruct
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> mi_file = magni.afm.io.read_mi_file(path)
    >>> mi_buffer = mi_file.get_buffer('Topography')[0]
    >>> mi_data = mi_buffer.get_data()
    >>> x = magni.imaging.mat2vec(mi_data)

    Next, a measurement matrix is defined. This matrix is equal to the matrix
    generated by running `np.eye(len(x))[::2, :]` but for speed, the matrix is
    instead defined with fast operations:

    >>> def Phi_A(x):
    ...     y = x[::2]
    ...     return y
    >>> def Phi_T(y):
    ...     x = np.zeros((2 * len(y), 1))
    ...     x[::2] = y
    ...     return x
    >>> Phi = magni.utils.matrices.Matrix(Phi_A, Phi_T, (),
    ...                                   (int(len(x) / 2), len(x)))

    Next, a dictionary is defined. This dictionary is the DCT basis likewise
    defined with fast operations:

    >>> Psi = magni.imaging.dictionaries.get_DCT(mi_data.shape)

    Finally, the actual example:

    >>> y = Phi.dot(x)
    >>> print('Maximum absolute pixel error: {:.3f}'
    ...       .format(np.abs(reconstruct(y, Phi, Psi) - x).max()))
    Maximum absolute pixel error: 0.960

    """

    _validate_reconstruct(y, Phi, Psi)

    A = _MatrixC((Phi, Psi))
    algorithm = _config.get('algorithm')

    if algorithm == 'iht':
        algorithm = _iht.run
    elif algorithm == 'sl0':
        A = A.A
        algorithm = _sl0.run

    alpha = algorithm(y, A)
    x = Psi.dot(alpha)
    return x



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public functions of the magni.imaging subpackage.

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate
from magni.utils.validation import validate_ndarray as _validate_ndarray


@_decorate_validation
def _validate_mat2vec(x):
    """
    Validatate the `mat2vec` function.

    See also
    --------
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(x, 'x', {'dim': 2})


def mat2vec(x):
    """
    Reshape `x` from matrix to vector by stacking columns.

    Parameters
    ----------
    x : ndarray
        Matrix that should be reshaped to vector.

    Returns
    -------
    ndarray
        Column vector formed by stacking the columns of the matrix `x`.

    See Also
    --------
    vec2mat : The inverse operation

    Notes
    -----
    The returned column vector is C contiguous.

    Examples
    --------
    For example,

    >>> from magni.imaging._util import mat2vec
    >>> x = np.arange(4).reshape(2, 2)
    >>> x
    array([[0, 1],
           [2, 3]])
    >>> mat2vec(x)
    array([[0],
           [2],
           [1],
           [3]])

    """

    _validate_mat2vec(x)

    return x.T.reshape(-1, 1)


@_decorate_validation
def _validate_vec2mat(x, mn_tuple):
    """
    Validatate the `vec2mat` function.

    See also
    --------
    magni.utils.validation.validate : Validation.

    """

    m, n = mn_tuple

    _validate(m, 'm', {'type': int, 'min': 1})
    _validate(n, 'n', {'type': int, 'min': 1})
    _validate_ndarray(x, 'x', {'shape': (m * n, 1)})


def vec2mat(x, mn_tuple):
    """
    Reshape `x` from column vector to matrix.

    Parameters
    ----------
    x : ndarray
        Matrix that should be reshaped to vector.
    mn_tuple : tuple
        A tuple (m, n) containing the parameters m, n as listed below.
    m : int
        Number of rows in the resulting matrix.
    n : int
        Number of columns in the resulting matrix.

    Returns
    -------
    ndarray
       Matrix formed by taking `n` columns of lenght `m` from the column vector
       `x`.

    See Also
    --------
    mat2vec : The inverse operation

    Notes
    -----
    The returned matrix is C contiguous.

    Examples
    --------
    For example,

    >>> from magni.imaging._util import vec2mat
    >>> x = np.arange(4).reshape(4, 1)
    >>> x
    array([[0],
           [1],
           [2],
           [3]])
    >>> vec2mat(x, (2, 2))
    array([[0, 2],
           [1, 3]])

    """

    m, n = mn_tuple

    _validate_vec2mat(x, (m, n))

    return x.reshape(n, m).T



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions that may be used to annotate data.

Routine Listings
----------------
get_conda_info()
    Function that returns information about a Continnum Anaconda install.
get_datetime()
    Function that returns information about the current date and time.
get_git_revision()
    Function that returns information about the `magni` git revision.
get_magni_config()
    Function that returns information about the current configuration of Magni.
get_magni_info()
    Function that returns genral information about Magni.
get_platform_info()
    Function that returns information about the platform used to run the code.

Notes
-----
The return annotations are any nested level of dicts of dicts of strings.

"""

from __future__ import division
import datetime
import json
import os
import pkgutil
import platform
import pydoc
import re
import subprocess

import magni


def get_conda_info():
    """
    Return a dictionary contianing information from Conda.

    Conda is the package manager for the `Continuum Anaconda
    <https://store.continuum.io/cshop/anaconda/>`_ scientific Python
    distribution. This function will return various information about the
    Anaconda installation on the system by quering the Conda package database.

    Returns
    -------
    conda_info : dict
        Various information from conda (see notes below for further details).

    Notes
    -----
    If the Python intepreter is unable to locate and import the conda package,
    an empty dicionary is returned.

    The returned dictionary contains the same infomation that is returned by
    "conda info" in addition to an overview of the linked modules in the
    Anaconda installation. Specifically, the returned dictionary has the
    following keys:

    * platform
    * conda_version
    * root_prefix
    * default_prefix
    * envs_dirs
    * package_cache
    * channels
    * config_file
    * is_foreign_system
    * linked_modules

    Additionally, the returned dictionary has a key named *status*, which can
    have either of the following values:

    * 'Succeeded' (Everything seems to be OK)
    * 'Failed' (Import of conda failed - nothing else is returned)

    """

    try:
        import conda
        import conda.config
        import conda.install
    except ImportError:
        return {'status': 'Failed'}

    conda_info = {'platform': conda.config.subdir,
                  'conda_version': conda.__version__,
                  'root_prefix': conda.config.root_dir,
                  'default_prefix': conda.config.default_prefix,
                  'envs_dirs': json.dumps(conda.config.envs_dirs),
                  'package_cache': json.dumps(conda.config.pkgs_dirs),
                  'channels': json.dumps(conda.config.get_channel_urls()),
                  'config_file': json.dumps(conda.config.rc_path),
                  'is_foreign_system': json.dumps(bool(conda.config.foreign)),
                  'linked_modules': sorted(
                      conda.install.linked(conda.config.root_dir))}

    modules_info = {module:
                    conda.install.is_linked(conda_info['root_prefix'], module)
                    for module in conda_info['linked_modules']}
    conda_info['modules_info'] = modules_info
    conda_info['linked_modules'] = json.dumps(conda_info['linked_modules'])

    conda_info['status'] = 'Succeeded'

    return conda_info


def get_datetime():
    """
    Return a dictionary holding the current date and time.

    Returns
    -------
    date_time : dict
        The dictionary holding the current date and time.

    Notes
    -----
    The returned dictionary has the following keys:

    * today (date and time including timezone offset)
    * utcnow (UTC date and time)
    * pretty_utc (UTC date and time formatted according to current locale)
    * status

    The status entry informs about the success of the pretty_utc formatting.
    It has one of the follwing values:

    * Succeeded (Everything seems OK)
    * Failed (It was not possible to format the time)

    """

    date_time = {'today': repr(datetime.datetime.today()),
                 'utcnow': datetime.datetime.utcnow(),
                 'pretty_utc': '',
                 'status': 'Succeeded'}

    try:
        date_time['pretty_utc'] = datetime.datetime.strftime(
            date_time['utcnow'], '%c')

    except ValueError:
        date_time['status'] = 'Failed'

    date_time['utcnow'] = repr(date_time['utcnow'])

    return date_time


def get_git_revision():
    """
    Return a dictionary containing information about the current git revision.

    Returns
    -------
    git_revision : dict
        Information about the current git revision.

    Notes
    -----
    If the git revision extract succeeded, the returned dictionary has the
    following keys:

    * status (with value 'Succeeded')
    * tag (output of "git describe")
    * branch (output of "git describe --all")

    If the git revision extract failed, the returned dictionary has the
    following keys:

    * status (with value 'Failed')
    * returncode (returncode from failing git command)
    * output (output from failing git command)

    The "git describe" commands are run in the directory in which `magni` is
    loaded from.

    """

    cur_dir = os.getcwd()
    os.chdir(os.path.split(magni.__path__[0])[0])

    try:
        git_revision = {'tag': str(subprocess.check_output(
            ['git', 'describe'], stderr=subprocess.STDOUT)[:-1].decode()),
            'branch': str(subprocess.check_output(
                ['git', 'describe', '--all'],
                stderr=subprocess.STDOUT)[:-1].decode()),
            'status': 'Succeeded'}

    except subprocess.CalledProcessError as e:
        git_revision = {'status': 'Failed: CallProcessError',
                        'returncode': e.returncode,
                        'output': e.output}

    except OSError as e:
        git_revision = {'status': 'Failed: OSError',
                        'errno': e.errno,
                        'strrror': e.strerror}

    os.chdir(cur_dir)

    return git_revision


def get_magni_config():
    """
    Return a dictionary holding the current configuration of Magni.

    Returns
    -------
    magni_config : dict
        The dictionary holding the current configuration of Magni.

    Notes
    -----
    The returned dictionary has a key for each of the `config` modules in Magni
    and its subpackages. The value of a given key is a dictionary with the
    current configuration of the corresponding `config` module. Furthermore,
    the returned dictionary has a status key, which can have either of the
    following values:

    * Succeeded (The entire configuration was extracted)
    * Failed (It was not possible to get information from one or more modules)

    """

    packages = pkgutil.walk_packages(path=magni.__path__,
                                     prefix=magni.__name__ + '.')

    magni_config = {'status': 'Succeeded'}
    try:
        for importer, modname, ispkg in packages:
            if modname[-7:] == '.config' and modname != 'magni.utils.config':
                settings = eval(modname + '.get()')
                for setting in settings:
                    if not isinstance(settings[setting], str):
                        settings[setting] = repr(settings[setting])

                magni_config[modname] = settings

    except AttributeError:
        magni_config['status'] = 'Failed'

    return magni_config


def get_magni_info():
    """
    Return a string representation of the output of help(magni).

    Returns
    -------
    magni_info : dict
        Information about magni.

    Notes
    -----
    The returned dictionary has a single key:

    * help_magni (a string representation of help(magni))

    """

    magni_info = pydoc.render_doc(magni)
    magni_info, subs = re.subn(r'\x08([A-Z]|[a-z]|_)?', '', magni_info)

    return {'help_magni': magni_info}


def get_platform_info():
    """
    Return a dictionary containing information about the system platform.

    Returns
    -------
    platform_info : dict
        Various information about the system platform.

    See Also
    --------
    platform : The Python module used to query information about the system.

    Notes
    -----
    The returned dictionary has the following keys:

    * system
    * node
    * release
    * version
    * processor
    * python
    * libc
    * linux
    * mac_os
    * win32
    * status

    The linux/mac_os/win32 entries are "empty" if they are not applicable.

    If the processor information returned by `platform` is "empty", a query of
    `lscpu` is attempted in order to provide the necessary information.

    The status entry informs about the success of the queries. It has one of
    the follwing values:

    * 'All OK' (everything seems to be OK)
    * 'Used lscpu in processor query' (`lscpu` was used)
    * 'Processor query failed' (failed to get processor information)

    """

    platform_info = {'system': json.dumps(platform.system()),
                     'node': json.dumps(platform.node()),
                     'release': json.dumps(platform.release()),
                     'version': json.dumps(platform.version()),
                     'machine': json.dumps(platform.machine()),
                     'processor': json.dumps(platform.processor()),
                     'python': json.dumps(platform.python_version()),
                     'libc': json.dumps(platform.libc_ver()),
                     'linux': json.dumps(platform.linux_distribution()),
                     'mac_os': json.dumps(platform.mac_ver()),
                     'win32': json.dumps(platform.win32_ver()),
                     'status': 'All OK'}

    if platform_info['processor'] == '':
        try:
            platform_info['processor'] = str(
                subprocess.check_output(['lscpu']).decode())
            platform_info['status'] = 'Used lscpu in processor query'

        except (subprocess.CalledProcessError, OSError):
            platform_info['status'] = 'Processor query failed'

    return platform_info



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing input/output functionality for MI files.

Routine listings
----------------
read_mi_file(path)
    Read MI file and output an instance of an appropriate class.

See Also
--------
magni.afm.types : Data container classes.

"""


from __future__ import division
import os

import numpy as np

from magni.afm import types as _types
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate


@_decorate_validation
def _validate_read_mi_file(path):
    """
    Validate the `read_mi_file` function.

    See Also
    --------
    read_mi_file : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(path, path, {'type': str})


def read_mi_file(path):
    """
    Read MI file and output an instance of an appropriate class.

    Parameters
    ----------
    path : str
        The path of the MI file.

    Returns
    -------
    obj : None
        An instance of an appropriate class depending on the content of the MI
        file.

    Notes
    -----
    See the specification of the MI file format for an understanding of the
    steps performed in reading the MI file. An MI file can contain different
    types of data and thus the class of the output can vary.

    Examples
    --------
    An example of how to use read_mi_file to read the example MI file provided
    with the package:

    >>> from magni.afm.io import read_mi_file
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> mi_file = read_mi_file(path)

    """

    _validate_read_mi_file(path)

    if not os.path.isfile(path):
        raise IOError("The file {!r} does not exist.".format(path))

    try:
        f = open(path, 'rb')
        buf = f.read()
    finally:
        f.close()

    index = buf.find(b'data')
    index = buf.find(b'\n', index)
    hdrs = buf[:index].split(b'\n')
    hdrs = [(str(hdr[:14].decode()), str(hdr[14:].decode())) for hdr in hdrs]
    hdrs = [(key.strip(), _convert_mi_value(val)) for key, val in hdrs]

    if hdrs[0][0].lower() != 'filetype' or hdrs[-1][0].lower() != 'data':
        raise IOError("The file {!r} is not a valid MI file.".format(path))

    buf = buf[index + 1:]

    if hdrs[0][1].lower() == 'image':
        data = _convert_mi_image_data(buf, hdrs[-1][1])
        obj = _types.Image(data, hdrs)
    elif hdrs[0][1].lower() == 'spectroscopy':
        raise NotImplementedError("'Spectroscopy' file type not implemented.")
    else:
        raise IOError("Unknown file type {!r}.".format(hdrs[0][1]))

    return obj


def _convert_mi_image_data(buf, datatype):
    """
    Convert the data part of an MI image file to a 1D numpy.ndarray.

    Parameters
    ----------
    buf : str
        The raw data part of an MI image file.
    datatype : str
        A string specifying how to interpret the data in the data buffer.

    Returns
    -------
    data : numpy.ndarray
        The converted data part.

    Notes
    -----
    See the specification of the MI file format for a list of datatypes and an
    explanation of their meaning.

    """

    if datatype == 'BINARY':
        data = np.frombuffer(buf, np.int16)
        data = np.float64(data) / 2**15
    elif datatype == 'BINARY_32':
        data = np.frombuffer(buf, np.int32)
        data = np.float64(data) / 2**31
    elif datatype == 'ASCII':
        data = np.int16([int(val) for val in buf.split()])
        data = np.float64(data) / 2**15
    elif datatype == 'ASCII_ABSOLUTE':
        data = np.float64([float(val) for val in buf.split()])
    elif datatype == 'ASCII_MULTICOLUMN':
        data = np.float64([[float(val) for val in row.split()]
                           for row in buf.split('\n')]).flatten('F')
    else:
        raise IOError("Unknown data type {!r}.".format(datatype))

    return data


def _convert_mi_value(string):
    """
    Convert the value of an MI header line to a meaningful Python type.

    Parameters
    ----------
    string : str
        The string representation of the MI header line value.

    Returns
    -------
    value : None
        The converted value.

    Notes
    -----
    See the specification of the MI file format for an explanation of the
    different value types and the conversion from the string representation.

    """

    parts = string.split()

    if len(parts) == 0:
        value = None
    elif len(parts) == 1:
        if string.lower() == 'true':
            value = True
        elif string.lower() == 'false':
            value = False
        else:
            try:
                value = string
                value = float(string)
                value = int(string)
            except ValueError:
                pass
    else:
        value = [_convert_mi_value(part) for part in parts]
        value = string if isinstance(value[0], str) else tuple(value)

    return value



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public function of the magni.cs.reconstruction.sl0
subpackage.

"""

from __future__ import division

from magni.cs.reconstruction.sl0 import config as _config
from magni.cs.reconstruction.sl0 import _modified
from magni.cs.reconstruction.sl0 import _original
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_ndarray as _validate_ndarray


@_decorate_validation
def _validate_run(y, A):
    """
    Validate the `run` function.

    See Also
    --------
    run : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(A, 'A', {'dim': 2})
    _validate_ndarray(y, 'y', {'shape': (A.shape[0], 1)})


def run(y, A):
    """
    Run the specified SL0 reconstruction algorithm.

    The available SL0 reconstruction algorithms are the original SL0 and the
    modified SL0. Which of the available SL0 reconstruction algorithms is used,
    is specified as a configuration option.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    See Also
    --------
    magni.cs.reconstruction.sl0.config : Configuration options.
    magni.cs.reconstruction.sl0._original.run : The original SL0 reconstruction
        algorithm.
    magni.cs.reconstruction.sl0._modified.run : The modified SL0 reconstruction
        algorithm.

    Examples
    --------
    See the individual run functions in the implementations of the original and
    modified SL0 reconstruction algorithms.

    """

    _validate_run(y, A)

    algorithm = _config.get('algorithm')

    if algorithm == 'std':
        x = _original.run(y, A)
    elif algorithm == 'mod':
        x = _modified.run(y, A)

    return x



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for visualising images.

The module provides functionality for adjusting the intensity of an image.
Furthermore, it provides a wrapper of the `matplotlib.pyplot.imshow` function
that may exploit the provided functions for adjusting the image intensity.

Routine listings
----------------
imshow(X, ax=None, intensity_func=None, intensity_args=(), \*\*kwargs)
    Function that may be used to display an image.
shift_mean(x_mod, x_org)
    Function for shifting mean intensity of an image based on another image.
stretch_image(img, max_val)
    Function for stretching the intensity of an image.

"""

from __future__ import division
import types

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from magni.utils import plotting as _plotting
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate
from magni.utils.validation import validate_ndarray as _validate_ndarray


@_decorate_validation
def _validate_imshow(X, ax, intensity_func, intensity_args, show_axis):
    """
    Validate the `imshow` function.

    See Also
    --------
    imshow : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(X, 'X')
    _validate(ax, 'ax', {'class': mpl.axes.Axes}, ignore_none=True)
    _validate(intensity_func, 'intensity_func', {'type': types.FunctionType},
              ignore_none=True)
    _validate(intensity_args, 'intensity_args', {'type_in': (list, tuple)})
    _validate(show_axis, 'show_axis', {'type': str,
                                       'val_in': ('none', 'top', 'inherit')})


def imshow(X, ax=None, intensity_func=None, intensity_args=(),
           show_axis='none', **kwargs):
    """
    Display an image.

    Wrap `matplotlib.pyplot.imshow` to display a possibly intensity manipulated
    version of the image `X`.

    Parameters
    ----------
    X : ndarray
        The image to be displayed.
    ax : matplotlib.axes.Axes, optional
        The axes on which the image is displayed (the default is None, which
        implies that the current axes is used).
    intensity_func : FunctionType, optional
        The handle to the function used to manipulate the image intensity
        before the image is displayed (the default is None, which implies that
        no intensity manipulation is used).
    intensity_args : list or tuple, optional
        The arguments that are passed to the `intensity_func` (the default is
        (), which implies that no arguments are passed).
    show_axis : {'none', 'top', 'inherit'}
        How the x- and y-axis are display. If 'none', no axis are displayed. If
        'top', the x-axis is displayed at the top of the image. If 'inherit',
        the axis display is inherited from `matplotlib.pyplot.imshow`.

    Returns
    -------
    im_out : matplotlib.image.AxesImage
        The AxesImage returned by matplotlibs imshow.

    See Also
    --------
    matplotlib.pyplot.imshow : Matplotlib's imshow function.

    Examples
    --------
    For example,

    >>> from magni.imaging.visualisation import imshow
    >>> X = np.arange(4).reshape(2, 2)
    >>> add_k = lambda X, k: X + k
    >>> im_out = imshow(X, intensity_func=add_k, intensity_args=(2,))

    """

    _validate_imshow(X, ax, intensity_func, intensity_args, show_axis)

    _plotting.setup_matplotlib()

    # Handle ax keyword argument
    ca = plt.gca()
    if ax is not None:
        plt.sca(ax)

    # Intensity manipulation
    if intensity_func is not None:
        im_out = plt.imshow(intensity_func(X, *intensity_args), **kwargs)
    else:
        im_out = plt.imshow(X, **kwargs)

    # Display of axis
    axes = plt.gca()
    if show_axis == 'none':
        xlabels = axes.get_xticklabels()
        ylabels = axes.get_yticklabels()
        empty_xlabels = ['']*len(xlabels)
        empty_ylabels = ['']*len(ylabels)
        axes.set_xticklabels(empty_xlabels)
        axes.set_yticklabels(empty_ylabels)
        axes.tick_params(length=0)
    elif show_axis == 'top':
        axes.xaxis.tick_top()
        axes.xaxis.set_label_position('top')

    plt.sca(ca)

    return im_out


@_decorate_validation
def _validate_shift_mean(x_mod, x_org):
    """
    Validate the `shift_mean` function.

    See Also
    --------
    shift_mean : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(x_org, 'x_mod', {})
    _validate_ndarray(x_mod, 'x_org', {'dtype': x_mod.dtype,
                                       'shape': x_mod.shape})


def shift_mean(x_mod, x_org):
    """
    Shift the mean value of `x_mod` such that it equals the mean of `x_org`.

    Parameters
    ----------
    x_org : ndarray
        The array which hold the "true" mean value.
    x_mod : ndarray
        The modified copy of `x_org` which must have its mean value shifted.

    Returns
    -------
    shifted_x_mod : ndarray
        A copy of `x_mod` with the same mean value as `x_org`.

    Examples
    --------
    For example,

    >>> from magni.imaging.visualisation import shift_mean
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_mod = np.ones((2, 2))
    >>> x_org.mean()
    1.5
    >>> x_mod.mean()
    1.0
    >>> shifted_x_mod = shift_mean(x_mod, x_org)
    >>> shifted_x_mod.mean()
    1.5
    >>> shifted_x_mod
    array([[ 1.5,  1.5],
           [ 1.5,  1.5]])

    """

    _validate_shift_mean(x_mod, x_org)

    return x_mod + (x_org.mean() - x_mod.mean())


@_decorate_validation
def _validate_stretch_image(img, max_val):
    """
    Validate the `stretch_image` function.

    See Also
    --------
    stretch_image : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(img, 'img', {'subdtype': np.floating})
    _validate(max_val, 'max_val', {'type_in': [int, float], 'min': 0})


def stretch_image(img, max_val):
    """
    Stretch image such that pixels values are in the range [0, `max_val`].

    Parameters
    ----------
    img : ndarray
        The (float) image that is to be stretched.
    max_val : int or float
        The maximum value in the stretched image.

    Returns
    -------
    stretched_img : ndarray
        A stretched copy of the input image.

    Notes
    -----
    The pixel values in the input image are scaled to lie in the interval [0,
    `max_val`] using a linear stretch.

    Examples
    --------
    For example,

    >>> from magni.imaging.visualisation import stretch_image
    >>> img = np.arange(4, dtype=np.float).reshape(2, 2)
    >>> stretched_img = stretch_image(img, 1.0)
    >>> stretched_img
    array([[ 0.        ,  0.33333333],
           [ 0.66666667,  1.        ]])

    """

    _validate_stretch_image(img, max_val)

    min_ = img.min()
    max_ = img.max()

    return max_val / (max_ - min_) * (img - min_)



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing data container classes for MI files.

The classes of this module can be used either directly or indirectly through
the `io` module by loading an MI file.

Routine listings
----------------
Buffer()
    Data class for MI image buffer.
Image()
    Data class for MI image.

See Also
--------
magni.afm.io : MI file loading.

"""

# the name of the present module makes the built-in module 'types' unavailable
from __future__ import absolute_import
from __future__ import division
import types

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate
from magni.utils.validation import validate_ndarray as _validate_ndarray


class Buffer():
    """
    Data class for MI image buffer.

    This class contains both buffer specific header lines and the 2D data of
    the buffer.

    Parameters
    ----------
    data : numpy.ndarray
        The data of the buffer represented as a 1D numpy.ndarray.
    hdrs : list or tuple
        The buffer specific header lines.
    width : int
        The width in pixels of the area covered by the buffer.
    height : int
        The height in pixels of the area covered by the buffer.

    Notes
    -----
    See `_image_headers` for a description of the header lines.

    Examples
    --------
    The `__init__` function is implicitly called when loading, for example, the
    MI file provided with the package:

    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> mi_file = magni.afm.io.read_mi_file(path)
    >>> mi_buffer = mi_file.get_buffer()[0]

    This buffer can have a number of attributes (stored as header lines in the
    MI file) including the 'bufferLabel' attribute:

    >>> print(mi_buffer.get_attr('bufferLabel'))
    Topography

    The primary purpose of this class is, however, to contain the 2D data of a
    buffer:

    >>> data = mi_buffer.get_data()
    >>> print('Buffer, Type: {}, Shape: {}'.format(str(type(data))[-15:-2],
    ... data.shape))
    Buffer, Type: numpy.ndarray, Shape: (256, 256)

    """

    @_decorate_validation
    def _validate_init(self, data, hdrs, width, height):
        """
        Validate the `__init__` function.

        See Also
        --------
        Buffer.__init__ : The validated function.
        magni.utils.validation.validate : Validation.
        magni.utils.validation.validate_ndarray : Validation.

        """

        _validate_ndarray(data, 'data')
        _validate(hdrs, 'hdrs', {'type_in': (list, tuple)})
        _validate(width, 'width', {'type': int, 'min': 1})
        _validate(height, 'height', {'dtype': int, 'min': 1})

    def __init__(self, data, hdrs, width, height):
        self._validate_init(data, hdrs, width, height)

        self._headers = {key: val for key, val in hdrs if key in _image_keys}
        self._data = self.get_attr('bufferRange') * data
        self._data = self._data.reshape(height, width)
        self._data = self._data[::-1, :]

    @_decorate_validation
    def _validate_get_attr(self, key):
        """
        Validate the `get_attr` function.

        See Also
        --------
        Buffer.get_attr : The validated function.
        magni.utils.validation.validate : Validation.

        """

        _validate(key, 'key', {'val_in': list(self._headers.keys())}, True)

    def get_attr(self, key=None):
        """
        Get a copy of all header lines or a specific header line.

        Parameters
        ----------
        key : str or None, optional
            The name of the header line to retrieve (the default is None, which
            implies retrieving a copy of all header lines).

        Returns
        -------
        value : dict or None
            The value of the specified key, if key is not None. Otherwise, a
            copy of the header lines.

        """

        self._validate_get_attr(key)

        if key is None:
            value = self._headers.copy()
        else:
            value = self._headers[key]

        return value

    def _validate_get_data(self, intensity_func, intensity_args):
        """
        Validate the `get_data` function.

        See Also
        --------
        Buffer.get_data : The validated function.
        fat.utils.validation.validate : Validation.

        """

        _validate(intensity_func, 'intensity_func',
                  {'type': types.FunctionType}, ignore_none=True)
        _validate(intensity_args, 'intensity_args', {'type_in': (list, tuple)})

    def get_data(self, intensity_func=None, intensity_args=()):
        """
        Get the 2D data of the buffer.

        The optional `intensity_func` and `intensity_args` can be used to
        manipulate the intensity of the image before getting the data.

        Parameters
        ----------
        intensity_func : FunctionType, optional
            The handle to the function used to manipulate the image intensity
            (the default is None, which implies that no intensity manipulation
            is used).
        intensity_args : list or tuple, optional
            The arguments that are passed to the `intensity_func` (the default
            is (), which implies that no arguments are passed).

        Returns
        -------
        data : numpy.ndarray
            The 2D data of the buffer.

        """

        self._validate_get_data(intensity_func, intensity_args)

        data = self._data

        if intensity_func is not None:
            data = intensity_func(data, *intensity_args)

        return data


class Image():
    """
    Data class for MI image.

    This class contains both image specific header lines and the buffers of the
    image.

    Parameters
    ----------
    data : numpy.ndarray
        The data part of the MI image.
    hdrs : list or tuple
        The image specific header lines.

    Notes
    -----
    See `_mi_headers` for a description of the header lines.

    Examples
    --------
    The `__init__` function is implicitly called when loading, for example, the
    MI file provided with the package:

    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> image = magni.afm.io.read_mi_file(path)

    This image can have a number of attributes (stored as header lines in the
    MI file) including the 'scanSpeed' attribute:

    >>> print('{:5.2f}'.format(image.get_attr('scanSpeed')))
     1.01

    The primary purpose of this class is, however, to contain the buffers of an
    MI image file:

    >>> buffers = image.get_buffer()
    >>> for b in buffers[0:5:2]:
    ...     print('Buffer: {}'.format(b.get_attr('bufferLabel')))
    Buffer: Topography
    Buffer: Deflection
    Buffer: Friction

    """

    @_decorate_validation
    def _validate_init(self, data, hdrs):
        """
        Validate the `__init__` function.

        See Also
        --------
        Image.__init__ : The validated function.
        magni.utils.validation.validate : Validation.
        magni.utils.validation.validate_ndarray : Validation.

        """

        _validate_ndarray(data, 'data')
        _validate(hdrs, 'hdrs', {'type_in': (list, tuple)})

    def __init__(self, data, hdrs):
        self._validate_init(data, hdrs)

        end = [hdr[0] for hdr in hdrs].index(_image_keys[0])
        self._headers = {key: val for key, val in hdrs[:end] + [hdrs[-1]]
                         if key in _mi_keys}

        self._buffers = []
        width = self.get_attr('xPixels')
        height = self.get_attr('yPixels')
        size = width * height
        bufs = []

        for i in range(end, len(hdrs) - 1):
            if hdrs[i][0] == _image_keys[0]:
                bufs.append([])

            bufs[-1].append(hdrs[i])

        for i in range(len(bufs)):
            buf = data[i * size:(i + 1) * size]
            self._buffers.append(Buffer(buf, bufs[i], width, height))

    @_decorate_validation
    def _validate_get_attr(self, key):
        """
        Validate the `get_attr` function.

        See Also
        --------
        Image.get_attr : The validated function.
        magni.utils.validation.validate : Validation.

        """

        _validate(key, 'key', {'val_in': list(self._headers.keys())}, True)

    def get_attr(self, key=None):
        """
        Get a copy of all header lines or a specific header line.

        Parameters
        ----------
        key : str or None, optional
            The name of the header line to retrieve (the default is None, which
            implies retrieving a copy of all header lines).

        Returns
        -------
        value : dict or None
            The value of the specified key, if key is not None. Otherwise, a
            copy of the header lines.

        """

        self._validate_get_attr(key)

        if key is None:
            value = self._headers.copy()
        else:
            value = self._headers[key]

        return value

    @_decorate_validation
    def _validate_get_buffer(self, key):
        """
        Validate the `get_buffer` function.

        See Also
        --------
        Image.get_buffer : The validated function.
        magni.utils.validation.validate : Validation.

        """

        keys = [buffer.get_attr('bufferLabel') for buffer in self._buffers]
        _validate(key, 'key', {'val_in': keys}, True)

    def get_buffer(self, key=None):
        """
        Get all buffers or a specific buffer.

        Parameters
        ----------
        key : str or None, optional
            The name of the buffer to retrieve (the default is None, which
            implies retrieving all buffers).

        Returns
        -------
        value : dict or None
            The buffer of the specified key, if key is not None. Otherwise, a
            dict with all the buffers.

        """

        self._validate_get_buffer(key)

        if key is None:
            buf = self._buffers[:]
        else:
            buf = [buffer for buffer in self._buffers
                   if buffer.get_attr('bufferLabel') == key]

            if len(buf) == 1:
                buf = buf[0]

        return buf


_mi_headers = (
    ('fileType      ', 'Specifies if it is Image or plot file'),
    ('version       ', 'Software version'),
    ('dateAcquired  ', 'Date and time scan started'),
    ('dateModified  ', 'Date and time data saved'),
    ('comment       ', ''),
    ('mode          ', 'ModeAFM = Acquired in contact AFM'),
    ('xSensitivity  ', 'X sensitivity for scanner (m/V)'),
    ('xNonlinearity ', 'X non-linearity for scanner'),
    ('xHysteresis   ', 'X hysteresis for scanner (% scan size)'),
    ('ySensitivity  ', 'Y sensitivity for scanner (m/V)'),
    ('yNonlinearity ', 'Y non-linearity for scanner'),
    ('yHysteresis   ', 'Y hysteresis for scanner (% scan size)'),
    ('zSensitivity  ', 'Z sensitivity for scanner (m/V)'),
    ('reverseX      ', 'Reverse X polarity for scanner'),
    ('reverseY      ', 'Reverse Y polarity for scanner'),
    ('reverseZ      ', 'Reverse Z polarity for scanner'),
    ('xDacRange     ', 'X output range for controller (V)'),
    ('yDacRange     ', 'Y output range for controller (V)'),
    ('zDacRange     ', 'Z output range for controller (V)'),
    ('xPixels       ', 'Columns in image'),
    ('yPixels       ', 'Rows in image'),
    ('xOffset       ', 'X offset in total scan area'),
    ('yOffset       ', 'Y offset in total scan area'),
    ('xLength       ', 'X scan Size (m)'),
    ('yLength       ', 'Y scan Size (m)'),
    ('scanUp        ', 'Scanning up = TRUE, down = FALSE'),
    ('scanSpeed     ', 'Fast scan speed (lines/sec or Hz)'),
    ('scanAngle     ', 'Scan angle (degrees)'),
    ('servoSetpoint ', 'Z servo setpoint (V)'),
    ('biasSample    ', 'Bias applied to sample (TRUE), tip (FALSE)'),
    ('bias          ', 'Bias voltage (V)'),
    ('servoIGain    ', 'Z servo integral gain'),
    ('servoPGain    ', 'Z servo proportional gain'),
    ('servoRange    ', 'Z servo range (m)'),
    ('servoInputGain', 'Z servo input gain'),
    ('acFrequency   ', 'AC Mode frequency (Hz)'),
    ('acDrive       ', 'AC Mode drive (%)'),
    ('acGain        ', 'AC Mode input gain'),
    ('acMac         ', 'AC Mode MAC (TRUE), AAC (FALSE)'),
    ('acACMode      ', 'AC Mode AC (TRUE), contact (FALSE)'),
    ('snapFileName  ', 'Names of snap files'),
    ('snapAfmPos    ', 'XY-positions of snap images (m)'),
    ('snapAfmOpacity', 'Snap opacities in range [0, 100]%'),
    ('snapResolution', 'Snap resolutions (pixels)'),
    ('snapSens      ', 'Optical X and Y snap sensitivities (m/pixel)'),
    ('snapROI       ', 'ROI for each snap image'),
    ('snapObjective ', 'Objective names for each snap image'),
    ('snapName      ', 'Snap names'),
    ('tipX          ', 'Tip X position'),
    ('tipY          ', 'Tip Y position'),
    ('data          ', 'Data'))
_mi_headers = tuple((key.strip(), descr) for key, descr in _mi_headers)
_mi_keys = [hdr[0] for hdr in _mi_headers]

_image_headers = (
    ('bufferLabel   ', 'First buffer label'),
    ('trace         ', 'First buffer trace (TRUE), retrace (FALSE)'),
    ('bufferUnit    ', 'First buffer unit'),
    ('bufferRange   ', 'First buffer range (bufferUnit unit)'))
_image_headers = tuple((key.strip(), descr) for key, descr in _image_headers)
_image_keys = [hdr[0] for hdr in _image_headers]



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing fast linear operations wrapped in matrix emulators.

Routine listings
----------------
get_DCT(shape)
    Get the DCT fast operation dictionary for the given image shape.
get_DFT(shape)
    Get the DFT fast operation dictionary for the given image shape.

See Also
--------
magni.imaging._fastops : Fast linear operations.
magni.utils.matrices : Matrix emulators.

"""

from __future__ import division

from magni.imaging import _fastops
from magni.utils.matrices import Matrix as _Matrix
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate


@_decorate_validation
def _validate_get_DCT(shape):
    """
    Validate the `get_DCT` function.

    See Also
    --------
    get_DCT : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(shape, 'shape',
              [{'type_in': (list, tuple), 'len': 2}, {'type': int, 'min': 1}])


def get_DCT(shape):
    """
    Get the DCT fast operation dictionary for the given image shape.

    Parameters
    ----------
    shape : list or tuple
        The shape of the image which the dictionary is the DCT dictionary.

    Returns
    -------
    matrix : magni.utils.matrices.Matrix
        The specified DCT dictionary.

    See Also
    --------
    magni.utils.matrices.Matrix : The matrix emulator class.

    Examples
    --------
    Create a dummy image:

    >>> img = np.random.randn(64, 64)
    >>> vec = magni.imaging.mat2vec(img)

    Perform DCT in the ordinary way:

    >>> dct_normal = magni.imaging._fastops.dct2(vec, *img.shape)

    Perform DCT using the present function:

    >>> from magni.imaging.dictionaries import get_DCT
    >>> matrix = get_DCT(img.shape)
    >>> dct_matrix = matrix.T.dot(vec)

    Check that the two ways produce the same result:

    >>> np.allclose(dct_matrix, dct_normal)
    True

    """

    _validate_get_DCT(shape)

    entries = shape[0] * shape[1]
    return _Matrix(_fastops.idct2, _fastops.dct2, shape, (entries, entries))


@_decorate_validation
def _validate_get_DFT(shape):
    """
    Validate the `get_DFT` function.

    See Also
    --------
    get_DFT : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(shape, 'shape',
              [{'type_in': (list, tuple), 'len': 2}, {'type': int, 'min': 1}])


def get_DFT(shape):
    """
    Get the DFT fast operation dictionary for the given image shape.

    Parameters
    ----------
    shape : list or tuple
        The shape of the image which the dictionary is the DFT dictionary.

    Returns
    -------
    matrix : magni.utils.matrices.Matrix
        The specified DFT dictionary.

    See Also
    --------
    magni.utils.matrices.Matrix : The matrix emulator class.

    Examples
    --------
    Create a dummy image:

    >>> img = np.random.randn(64, 64)
    >>> vec = magni.imaging.mat2vec(img)

    Perform DFT in the ordinary way:

    >>> dft_normal = magni.imaging._fastops.dft2(vec, *img.shape)

    Perform DFT using the present function:

    >>> from magni.imaging.dictionaries import get_DFT
    >>> matrix = get_DFT(img.shape)
    >>> dft_matrix = matrix.T.dot(vec)

    Check that the two ways produce the same result:

    >>> np.allclose(dft_matrix, dft_normal)
    True

    """

    _validate_get_DFT(shape)

    entries = shape[0] * shape[1]
    return _Matrix(_fastops.idft2, _fastops.dft2, shape, (entries, entries))



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing a robust configger class.

Routine listings
----------------
Configger()
    Provide set and get functions to access a set of configuration options.

Notes
-----
This module does not itself contain any configuration options and thus has no
get or set functions unlike the other config modules of `magni`.

"""

from __future__ import division
from itertools import chain

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate


class Configger():
    """
    Provide set and get functions to access a set of configuration options.

    The set of configuration options, their default values, and their
    validation schemes are specified upon initialisation.

    Parameters
    ----------
    param : dict
        The configuration options along with their default values.
    requirements : dict
        The validation schemes of the configuration options.

    See Also
    --------
    magni.utils.validation : Validation.

    Notes
    -----
    `requirements` must contain the same keys as `param`. For each key in
    `requirements`, the value is used as validation scheme - see `set` for
    further information.

    Examples
    --------
    Instantiate Configger with the parameter 'key' with default value 'default'
    which can only assume string values.

    >>> from magni.utils.config import Configger
    >>> config = Configger({'key': 'default'}, {'key': {'type': str}})

    The parameter can either be retrieved by getting a copy of the entire
    parameter dictionary or by getting the specific key.

    >>> config.get()
    {'key': 'default'}
    >>> config.get('key')
    'default'

    Likewise the parameter can either be changed by passing a dictionary with
    the parameter or by using a keyword argument.

    >>> config.set({'key': 'value'})
    >>> config.set(key='value')
    >>> config.get('key')
    'value'

    """

    @_decorate_validation
    def _validate_init(self, param, requirements):
        """
        Validate the `__init__` function.

        See Also
        --------
        Configger.__init__ : The validated function.
        magni.utils.validation.validate : Validation.

        """

        _validate(param, 'param', {'type': dict})
        _validate(requirements, 'requirements',
                  [{'type': dict}, {'type_in': [list, tuple, dict]}])

        if not sorted(param.keys()) == sorted(requirements.keys()):
            raise KeyError('param and requirements must have the same keys.')

    def __init__(self, param, requirements):
        self._validate_init(param, requirements)

        self._param = param
        self._requirements = requirements

    @_decorate_validation
    def _validate_get(self, key):
        """
        Validate the `get` function.

        See Also
        --------
        Configger.get : The validated function.
        magni.utils.validation.validate : Validation.

        """

        _validate(key, 'key', {'val_in': list(self._param.keys())}, True)

    def get(self, key=None):
        """
        Retrieve a copy of all parameters or a specific parameter.

        Parameters
        ----------
        key : str or None, optional
            The name of the key to retrieve (the default is None, which implies
            retrieving a copy of all parameters)

        Returns
        -------
        value : dict or None
            The value of the specified key, if a key is not None. Otherwise,
            a copy of the parameter dictionary.

        """

        self._validate_get(key)

        if key is None:
            value = self._param.copy()
        else:
            value = self._param[key]

        return value

    @_decorate_validation
    def _validate_set(self, dictionary, kwargs):
        """
        Validate the `set` function.

        See Also
        --------
        Configger.set : The validated function.
        magni.utils.validation.validate : Validation.

        """

        keys = list(self._param.keys())
        _validate(dictionary, 'dictionary', {'type': dict, 'keys_in': keys})
        _validate(kwargs, 'kwargs', {'keys_in': keys})

        all_items = dict(chain(dictionary.items(), kwargs.items())).items()
        for key, value in all_items:
            _validate(value, key, self._requirements[key])

    def set(self, dictionary={}, **kwargs):
        """
        Overwrite the value of one or more parameters.

        Each value is validated according to the validation scheme of that
        parameter.

        Parameters
        ----------
        dictionary : dict, optional
            A dictionary containing the key and value pairs to update.
        kwargs : dict, optional
            Keyword arguments being the key and value pairs to update.

        See Also
        --------
        magni.utils.validation.validate : Validation.

        """

        self._validate_set(dictionary, kwargs)

        all_items = dict(chain(dictionary.items(), kwargs.items())).items()
        for key, value in all_items:
            self._param[key] = value



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the phase_transition subpackage.

Routine listings
----------------
get(key=None)
    Get the value of one or more configuration options.
set(dictionary={}, \*\*kwargs)
    Set the value of one or more configuration options.

See Also
--------
magni.utils.config : The Configger class used.

Notes
-----
This module instantiates the `Configger` class provided by `magni.utils.config`
and assigns handles for the `get` and `set` methods of that class instance. The
configuration options are the following:

seed : int
    The seed used when picking seeds for generating data for the monte carlo
    simulations (the default is None, which implies an arbitrary seed).
n : int
    The length of the coefficient vector (the default is 800).
delta : list or tuple
    The delta values of the delta-rho grid whose points are used for the monte
    carlo simulations (the default is [0., 1.]).
rho : list or tuple
    The rho values of the delta-rho grid whose points are used for the monte
    carlo simulations (the default is [0., 1.]).
monte_carlo : int
    The number of monte carlo simulations to run in each point of the delta-rho
    grid (the default is 1).
coefficients : {'rademacher', 'gaussian'}
    The distribution which the non-zero coefficients in the coefficient vector
    are drawn from.

"""

from __future__ import division

from magni.utils.config import Configger as _Configger

_configger = _Configger(
    {'seed': None, 'n': 800, 'delta': [0.0, 1.0], 'rho': [0.0, 1.0],
     'monte_carlo': 1, 'coefficients': 'rademacher'},
    {'seed': {'type': int},
     'n': {'type': int, 'min': 1},
     'delta': [{'type_in': (list, tuple)},
               {'type': float, 'min': 0.0, 'max': 1.0}],
     'rho': [{'type_in': (list, tuple)},
             {'type': float, 'min': 0.0, 'max': 1.0}],
     'monte_carlo': {'type': int, 'min': 1},
     'coefficients': {'type': str, 'val_in': ['rademacher', 'gaussian']}})

get = _configger.get
set = _configger.set



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for constructing scan patterns for measurements.

This module provides several pairs of scan pattern functions. The first
function, named \*_sample_surface, is used for sampling a given surface. The
second function, named \*_sample_image, is a wrapper that provides a
pixel-oriented interface to the first function. In a addition to these pairs of
scan pattern functions, the module provides auxillary functions that may be
used to visualise the scan patterns.

Routine listings
----------------
construct_measurement_matrix(coords, h, w)
    Function for constructing a measurement matrix.
plot_pattern(l, w, coords, mode, output_path=None)
    Function for visualising a scan pattern.
plot_pixel_mask(h, w, pixels, output_path=None)
    Function for visualising a pixel mask obtained from a scan pattern.
random_line_sample_image(h, w, scan_length, num_points, discrete=None,
    seed=None)
    Function for random line sampling an image.
random_line_sample_surface(l, w, speed, sample_rate, time, discrete=None,
    seed=None)
    Function for random line sampling a surface.
spiral_sample_image(h, w, scan_length, num_points)
    Function for spiral sampling an image.
spiral_sample_surface(l, w, speed, sample_rate, time)
    Function for spiral sampling a surface.
square_spiral_sample_image(h, w, scan_length, num_points)
    Function for square spiral sampling an image.
square_spiral_sample_surface(l, w, speed, sample_rate, time)
    Function for square spiral sampling a surface.
uniform_line_sample_image(h, w, scan_length, num_points)
    Function for uniform line sampling an image.
uniform_line_sample_surface(l, w, speed, sample_rate, time)
    Function for uniform line sampling a surface.
unique_pixels(coords)
    Function for determining unique pixels from a set of coordinates.

Notes
-----
In principle, most of the scan pattern related parameters need only be
positive. However, it is assumed that the following requirements are fulfilled:

:Minimum length of scan area: 1 nm
:Minimum width of scan area: 1 nm
:Minimum scan speed: 1 nm/s
:Minimum sample_rate: 1 Hz
:Minimum scan time: 1 s
:Minimum scan length: 1 nm
:Minimum number of scan points: 1

Examples
--------
Sample a surface using a spiral pattern:

>>> from magni.imaging.measurements import spiral_sample_surface
>>> l = 13.0; w = 13.0; speed = 4.0; time = 27.0; sample_rate = 3.0;
>>> coords = spiral_sample_surface(l, w, speed, sample_rate, time)

Display the resulting pattern:

>>> from magni.imaging.measurements import plot_pattern
>>> plot_pattern(l, w, coords, 'surface')

Sample a 128x128 pixel image using random lines and a fixed seed:

>>> from magni.imaging.measurements import random_line_sample_image
>>> h = 128; w = 128; scan_length = 1000.0; num_points = 200; seed=6021;
>>> coords = random_line_sample_image(h, w, scan_length, num_points, seed=seed)

Display the resulting pattern:

>>> plot_pattern(h, w, coords, 'image')

Find the corresponding unique pixels and plot the pixel mask:

>>> from magni.imaging.measurements import unique_pixels, plot_pixel_mask
>>> unique_pixels = unique_pixels(coords)
>>> plot_pixel_mask(h, w, unique_pixels)

"""

from __future__ import division

import numpy as np
import matplotlib.pyplot as plt

from magni.imaging.visualisation import imshow as _imshow
from magni.utils.matrices import Matrix as _Matrix
from magni.utils import plotting as _plotting
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate
from magni.utils.validation import validate_ndarray as _validate_ndarray

# In principle most of the AFM-scanning related parameters should just be
# positive, however we have settled for:
_min_l = 1e-9  # [m]
_min_w = 1e-9  # [m]
_min_speed = 1e-9  # [m/s]
_min_sample_rate = 1.0  # [Hz]
_min_time = 1.0  # [s]
_min_scan_length = 1e-9  # [m]
_min_num_points = 1  # []


@_decorate_validation
def _validate_construct_measurement_matrix(coords, h, w):
    """
    Validate the `construct_measurement_matrix` function.

    See Also
    --------
    construct_measurement_matrix : The validated function.
    magni.utils.validation.validate : Validation.
    magni.utils.validation.validate_ndarray : Validation.

    """

    _validate_ndarray(coords, 'coords', {'shape': (coords.shape[0], 2)})
    _validate(h, 'h', {'type': int, 'min': 1})
    _validate(w, 'w', {'type': int, 'min': 1})


def construct_measurement_matrix(coords, h, w):
    """
    Construct a measurement matrix extracting the specified measurements.

    Parameters
    ----------
    coords : ndarray
        The `k` floating point coordinates arranged into a 2D array where each
        row is a coordinate pair (x, y), such that `coords` has size `k` x 2.

    Returns
    -------
    Phi : magni.utils.matrices.Matrix
        The constructed measurement matrix.

    See Also
    --------
    magni.utils.matrices.Matrix : The matrix emulator class.

    Notes
    -----
    The function construct two functions: one for extracting pixels at the
    coordinates specified and one for the transposed operation. These functions
    are then wrapped by a matrix emulator which is returned.

    Examples
    --------
    Create a dummy 5 by 5 pixel image and an example sampling pattern:

    >>> img = np.arange(25, dtype=np.float).reshape(5, 5)
    >>> vec = magni.imaging.mat2vec(img)
    >>> coords = magni.imaging.measurements.uniform_line_sample_image(
    ...              5, 5, 16., 17)

    Sample the image in the ordinary way:

    >>> unique = magni.imaging.measurements.unique_pixels(coords)
    >>> samples_normal = img[unique[:, 1], unique[:, 0]]
    >>> samples_normal = samples_normal.reshape((len(unique), 1))

    Sample the image using the present function:

    >>> from magni.imaging.measurements import construct_measurement_matrix
    >>> matrix = construct_measurement_matrix(coords, *img.shape)
    >>> samples_matrix = matrix.dot(vec)

    Check that the two ways produce the same result:

    >>> np.allclose(samples_matrix, samples_normal)
    True

    """

    _validate_construct_measurement_matrix(coords, h, w)

    coords = unique_pixels(coords)
    mask = coords[:, 0] * w + coords[:, 1]

    def measure(vec):
        return vec[mask]

    def measure_T(vec):
        output = np.zeros((h * w, 1))
        output[mask] = vec
        return output

    return _Matrix(measure, measure_T, [], (len(mask), h * w))


@_decorate_validation
def _validate_plot_pattern(l, w, coords, mode, output_path):
    """
    Validate the `plot_pattern` function.

    See Also
    --------
    plot_pattern : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(mode, 'mode', {'type': str, 'val_in': {'surface', 'image'}})
    _validate_ndarray(coords, 'coords', {'subdtype': np.floating, 'dim': 2})
    _validate(output_path, 'output_path', {'type': str}, ignore_none=True)

    if coords.shape[1] != 2:
        raise TypeError('coords.shape[1] must be 2.')

    if mode == 'surface':
        _validate(l, 'l', {'type': float, 'min': _min_l})
        _validate(w, 'w', {'type': float, 'min': _min_w})
    elif mode == 'image':
        _validate(l, 'l', {'type': int, 'min': 2})
        _validate(w, 'w', {'type': int, 'min': 2})


def plot_pattern(l, w, coords, mode, output_path=None):
    """
    Display a plot that shows the pattern given by a set of coordinates.

    The pattern given by the `coords` is displayed on an `w` x `l` area. If
    `mode` is 'surface', `l` and `w` are regarded as measured in meters. If
    `mode` is 'image', `l` and `w` are regarded as measured in pixels. The
    `coords` are marked by filled circles and connected by straight dashed
    lines.

    Parameters
    ----------
    l : float or int
        The length/height of the area. If `mode` is 'surface', it must be a
        float. If `mode` is 'image', it must be an integer.
    w : float or int
        The width of the area. If `mode` is 'surface', it must be a
        float. If `mode` is 'image', it must be an integer.
    coords : ndarray
        The 2D array of pixels that make up the mask. Each row is a coordinate
        pair (x, y).
    mode : {'surface', 'image'}
        The display mode that dertermines the axis labeling and the type of `l`
        and `w`.
    output_path : str, optional
        Path (including file type extension) under which the plot is saved (the
        default value is None which implies, that the plot is not saved).

    Notes
    -----
    The resulting plot is displayed in a figure using `matplotlib`'s
    `pyplot.plot`.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import plot_pattern
    >>> l = 3
    >>> w = 3
    >>> coords = np.array([[0, 0], [1, 1], [2, 1]], dtype=np.float)
    >>> mode = 'image'
    >>> plot_pattern(l, w, coords, mode)

    """

    _validate_plot_pattern(l, w, coords, mode, output_path)

    fs = (np.abs([l, w])).min() / 10
    _plotting.setup_matplotlib({'figure': {'figsize': (l / fs, w / fs)}})

    fig, axes = plt.subplots(1, 1)
    axes.plot(coords[:, 0], coords[:, 1],
              marker=_plotting.markers[0],
              ls=_plotting.linestyles[1],
              ms=6)

    axes.set_xlim([0, w])
    axes.xaxis.tick_top()
    axes.invert_yaxis()
    axes.set_ylim([l, 0])
    axes.xaxis.set_label_position('top')
    axes.set_aspect('equal')

    if mode == 'surface':
        axes.set_xlabel('width [m]')
        axes.set_ylabel('length [m]')
        axes.ticklabel_format(style='sci', scilimits=(0, 0))

        # Fix position of exponential multipliers in scientific notation
        exp_texts = [axes.xaxis.get_children()[1],
                     axes.yaxis.get_children()[1]]
        vert_aligns = ['bottom', 'center']
        hori_aligns = ['center', 'right']
        xy_texts = [(0, 25), (-35, 0)]
        xys = [(1, 1), (0, 0)]
        annos = [None, None]
        for k, text in enumerate(exp_texts):
            text.set_visible(False)
            annos[k] = axes.annotate('', xys[k], xytext=xy_texts[k],
                                     xycoords='axes fraction',
                                     textcoords='offset points',
                                     va=vert_aligns[k], ha=hori_aligns[k])

        fig.canvas.mpl_connect(
            'draw_event', lambda event: annos[0].set_text(
                exp_texts[0].get_text()))
        fig.canvas.mpl_connect(
            'draw_event', lambda event: annos[1].set_text(
                exp_texts[1].get_text()))

    elif mode == 'image':
        axes.set_xlabel('width [pixels]')
        axes.set_ylabel('height [pixels]')

    plt.tight_layout()

    if output_path is not None:
        plt.savefig(output_path)


@_decorate_validation
def _validate_plot_pixel_mask(h, w, pixels, output_path):
    """
    Validate the `plot_pixel_mask` function.

    See Also
    --------
    plot_pixel_mask : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h, 'h', {'type': int, 'min': 2})
    _validate(w, 'w', {'type': int, 'min': 2})
    _validate_ndarray(pixels, 'pixels', {'subdtype': np.integer, 'dim': 2})
    _validate(output_path, 'output_path', {'type': str}, ignore_none=True)

    if pixels.shape[1] != 2:
        raise TypeError('pixels.shape[1] must be 2.')

    if pixels.min() < 0:
        raise TypeError('All pixel coordinates must be non-negative.')

    if pixels[:, 0].max() > w - 1:
        raise TypeError('Pixel x-coordinates must not exceed pixel width.')

    if pixels[:, 1].max() > h - 1:
        raise TypeError('Pixel y-coordinates must not exceed pixel heigth.')


def plot_pixel_mask(h, w, pixels, output_path=None):
    """
    Display a binary image that shows the given pixel mask.

    A black image with `w` x `h` pixels is created and the `pixels` are marked
    with white.

    Parameters
    ----------
    h : int
        The height of the image in pixels.
    w : int
        The width of the image in pixels.
    pixels : ndarray
        The 2D array of pixels that make up the mask. Each row is a coordinate
        pair (x, y), such that `coords` has size len(`pixels`) x 2.
    output_path : str, optional
        Path (including file type extension) under which the plot is saved (the
        default value is None which implies, that the plot is not saved).

    Notes
    -----
    The resulting image is displayed in a figure using
    `magni.imaging.visualisation.imshow`.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import plot_pixel_mask
    >>> h = 3
    >>> w = 3
    >>> pixels = np.array([[0, 0], [1, 1], [2, 1]])
    >>> plot_pixel_mask(h, w, pixels)

    """

    _validate_plot_pixel_mask(h, w, pixels, output_path)

    mask = np.zeros((h, w))
    mask[pixels[:, 1], pixels[:, 0]] = 1

    fs = (np.abs([h, w])).min() / 10
    _plotting.setup_matplotlib({'figure': {'figsize': (h / fs, w / fs)}},
                               cmap='gray')

    fig, axes = plt.subplots(1, 1)
    _imshow(mask, ax=axes, show_axis='top')

    axes.set_xlabel('width [pixels]')
    axes.set_ylabel('height [pixels]')

    plt.tight_layout()

    if output_path is not None:
        plt.savefig(output_path)


@_decorate_validation
def _validate_random_line_sample_image(h, w, scan_length, num_points, discrete,
                                       seed):
    """
    Validate the `random_line_sample_image` function.

    See Also
    --------
    random_line_sample_image : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h, 'h', {'type': int, 'min': 2})
    _validate(w, 'w', {'type': int, 'min': 2})
    _validate(scan_length, 'scan_length', {'type': float,
                                           'min': _min_scan_length})
    _validate(num_points, 'num_points', {'type': int, 'min': _min_num_points})
    _validate(discrete, 'discrete', {'type': int, 'min': 2},
              ignore_none=True)
    _validate(seed, 'seed', {'type': int, 'min': 0}, ignore_none=True)


def random_line_sample_image(h, w, scan_length, num_points, discrete=None,
                             seed=None):
    """
    Sample an image using a set of random straight lines.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a pattern based on a set of random straight lines
    are determined. The `scan_length` determines the length of the path scanned
    whereas `num_points` indicates the number of samples taken on that path. If
    `discrete` is set, it specifies the finite number of equally spaced lines
    from which the scan lines are be chosen at random. For reproducible
    results, the `seed` may be used to specify a fixed seed of the random
    number generator.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.
    discrete : int or None, optional
        The number of equally spaced lines from which the scan lines are chosen
        (the default is None, which implies that no discritisation is used).
    seed : int or None, optional
        The seed used for the random number generator (the defaul is None,
        which implies that the random number generator is not seeded).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `scan_length` implies this. The top and bottom lines of the image are
    always included in the scan.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import random_line_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> seed = 6021
    >>> random_line_sample_image(h, w, scan_length, num_points, seed=seed)
    array([[ 0.5       ,  0.5       ],
           [ 5.04545455,  0.5       ],
           [ 9.5       ,  0.59090909],
           [ 5.65110302,  1.28746666],
           [ 1.10564847,  1.28746666],
           [ 0.5       ,  5.22727273],
           [ 3.29121172,  6.98151556],
           [ 7.83666626,  6.98151556],
           [ 7.23606783,  7.59970419],
           [ 2.69061328,  7.59970419],
           [ 0.95454545,  9.5       ],
           [ 5.5       ,  9.5       ]])

    """

    _validate_random_line_sample_image(h, w, scan_length, num_points, discrete,
                                       seed)

    coords = random_line_sample_surface(float(h - 1), float(w - 1),
                                        scan_length, float(num_points - 1),
                                        1.0, discrete=discrete, seed=seed)
    coords = coords + 0.5

    return coords


@_decorate_validation
def _validate_random_line_sample_surface(l, w, speed, sample_rate, time,
                                         discrete, seed):
    """
    Validate the `random_line_sample_surface` function.

    See Also
    --------
    random_line_sample_surface : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(l, 'l', {'type': float, 'min': _min_l})
    _validate(w, 'w', {'type': float, 'min': _min_w})
    _validate(speed, 'speed', {'type': float, 'min': _min_speed})
    _validate(sample_rate, 'sample_rate', {'type': float,
                                           'min': _min_sample_rate})
    _validate(time, 'time', {'type': float, 'min': _min_time})

    _validate(discrete, 'discrete', {'type': int, 'min': 2},
              ignore_none=True)
    _validate(seed, 'seed', {'type': int, 'min': 0}, ignore_none=True)

    if (speed * time - 2 * w - l) / w <= -1:
        # Estimated number of lines in addition to top and bottom lines must
        # exceed -1 to avoid drawing a negative number of lines at random.
        raise ValueError('(speed * time - 2 * w - l) / w must exceed -1.' +
                         'The current value is {!r}.'
                         .format((speed * time - 2 * w - l) / w))


def random_line_sample_surface(l, w, speed, sample_rate, time, discrete=None,
                               seed=None):
    """
    Sample a surface area using a set of random straight lines.

    The coordinates (in units of meters) resulting from sampling an image of
    size `l` times `w` using a pattern based on a set of random straight lines
    are determined.  The scanned path is determined from the probe `speed` and
    the scan `time`. If `discrete` is set, it specifies the finite number of
    equally spaced lines from which the scan lines are be chosen at random.
    For reproducible results, the `seed` may be used to specify a fixed seed of
    the random number generator.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.
    discrete : int or None, optional
        The number of equally spaced lines from which the scan lines are chosen
        (the default is None, which implies that no discritisation is used).
    seed : int or None, optional
        The seed used for the random number generator (the defaul is None,
        which implies that the random number generator is not seeded).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `speed` and `time` implies this. The top and bottom lines of the image are
    always included in the scan and are not included in the `discrete` number
    of lines.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import random_line_sample_surface
    >>> l = 2e-6
    >>> w = 2e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> seed = 6021
    >>> random_line_sample_surface(l, w, speed, sample_rate, time, seed=seed)
    array([[  0.00000000e+00,   0.00000000e+00],
           [  7.00000000e-07,   0.00000000e+00],
           [  1.40000000e-06,   0.00000000e+00],
           [  2.00000000e-06,   1.00000000e-07],
           [  1.37499259e-06,   1.74992590e-07],
           [  6.74992590e-07,   1.74992590e-07],
           [  0.00000000e+00,   2.00000000e-07],
           [  0.00000000e+00,   9.00000000e-07],
           [  2.22879572e-08,   1.57771204e-06],
           [  7.22287957e-07,   1.57771204e-06],
           [  1.42228796e-06,   1.57771204e-06],
           [  2.00000000e-06,   1.70000000e-06],
           [  1.60000000e-06,   2.00000000e-06]])


    """

    _validate_random_line_sample_surface(l, w, speed, sample_rate, time,
                                         discrete, seed)

    samples = int(sample_rate * time) + 1
    scan_length = speed * time
    sample_dist = scan_length / (samples - 1)
    lines_scan_length = scan_length - 2 * w - l
    num_lines = np.int(np.ceil(lines_scan_length / w))

    if seed is not None:
        np.random.seed(seed)
    lines = np.zeros(num_lines + 2)
    lines[-1] = l
    if discrete is None:
        lines[1:-1] = np.sort(np.random.rand(num_lines) * l)
    else:
        possible_lines = l / (discrete + 1) * np.arange(1, discrete + 1)
        try:
            lines[1:-1] = np.sort(np.random.choice(possible_lines,
                                                   size=num_lines,
                                                   replace=False))
        except ValueError:
            raise ValueError('The number of Discrete lines must be large ' +
                             'enough to contain the entire scan path. With ' +
                             'the current settings, a minimun of '
                             '{!r} lines are required.'.format(num_lines))

    coords = _get_line_scan_coords(lines, samples, sample_dist, l, w)

    return coords


@_decorate_validation
def _validate_spiral_sample_image(h, w, scan_length, num_points):
    """
    Validate the `spiral_sample_image` function.

    See Also
    --------
    spiral_sample_image : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h, 'h', {'type': int, 'min': 2})
    _validate(w, 'w', {'type': int, 'min': 2})

    if h != w:
        raise ValueError('h must equal w for an archimedian spiral to make ' +
                         'sense.')

    _validate(scan_length, 'scan_length', {'type': float,
                                           'min': _min_scan_length})
    _validate(num_points, 'num_points', {'type': int, 'min': _min_num_points})


def spiral_sample_image(h, w, scan_length, num_points):
    """
    Sample an image using an archimedean spiral pattern.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using an archimedean spiral pattern are determined. The
    `scan_length` determines the length of the path scanned whereas
    `num_points` indicates the number of samples taken on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis. The width must equal the height for an archimedian spiral to make
    sense.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import spiral_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> spiral_sample_image(h, w, scan_length, num_points)
    array([[ 6.28776846,  5.17074073],
           [ 3.13304898,  5.24133767],
           [ 6.07293751,  2.93873701],
           [ 6.99638041,  6.80851189],
           [ 2.89868434,  7.16724999],
           [ 2.35773914,  3.00320067],
           [ 6.41495385,  1.71018152],
           [ 8.82168896,  5.27557847],
           [ 6.34932919,  8.83624957],
           [ 2.04885699,  8.11199373],
           [ 0.6196052 ,  3.96939755]])

    """

    _validate_spiral_sample_image(h, w, scan_length, num_points)

    coords = spiral_sample_surface(float(h - 1), float(w - 1),
                                   scan_length, float(num_points), 1.0)
    coords = coords + 0.5

    return coords


@_decorate_validation
def _validate_spiral_sample_surface(l, w, speed, sample_rate, time):
    """
    Validate the `spiral_sample_surface` function.

    See Also
    --------
    spiral_sample_surface : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(l, 'l', {'type': float, 'min': _min_l})
    _validate(w, 'w', {'type': float, 'min': _min_w})

    if l != w:
        raise ValueError('l must equal w for an archimedian sprial to make' +
                         'sense.')

    _validate(speed, 'speed', {'type': float, 'min': _min_speed})
    _validate(sample_rate, 'sample_rate', {'type': float,
                                           'min': _min_sample_rate})
    _validate(time, 'time', {'type': float, 'min': _min_time})


def spiral_sample_surface(l, w, speed, sample_rate, time):
    """
    Sample a surface area using an archimedean spiral pattern.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using an archimedean spiral pattern are determined. The
    scanned path is determined from the probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis. The width must equal the length for an archimedian sprial to make
    sense.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import spiral_sample_surface
    >>> l = 1e-6
    >>> w = 1e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> spiral_sample_surface(l, w, speed, sample_rate, time)
    array([[  3.61074393e-07,   4.60846340e-07],
           [  5.16981905e-07,   7.08474495e-07],
           [  5.18627260e-07,   2.42473901e-07],
           [  5.90470818e-07,   7.85303986e-07],
           [  2.13777354e-07,   3.25138348e-07],
           [  8.38392713e-07,   3.55407335e-07],
           [  4.85640136e-07,   8.97652018e-07],
           [  9.79367276e-08,   3.60017412e-07],
           [  7.24145327e-07,   1.07666546e-07],
           [  8.91388500e-07,   7.71842556e-07],
           [  2.12224761e-07,   9.08883127e-07]])

    """

    _validate_spiral_sample_surface(l, w, speed, sample_rate, time)

    sample_period = 1 / sample_rate
    r_end = np.min([l, w]) / 2
    pitch = np.pi * r_end ** 2 / (speed * time)

    # Starting at t=0 yields a divide-by-zero problem. Therefore we start the
    # spiral at t=sample-period
    sample_times = np.linspace(sample_period, time, sample_rate * time - 1)

    r = np.sqrt(pitch * speed / np.pi * sample_times)
    theta = np.sqrt(4 * np.pi * speed * sample_times / pitch)

    coords = np.zeros((len(sample_times), 2))
    coords[:, 0] = r * np.cos(theta)
    coords[:, 1] = r * np.sin(theta)

    # FIXME: This simple stretch yields a non-constant linear speed, which
    # clashes with the idea of the current interface to all sample methods.
    """
    if w > l:
        coords[:, 0] = coords[:, 0] * w / l
    else:
        coords[:, 1] = coords[:, 1] * l / w
    """

    coords = coords + np.array([w / 2, l / 2])

    return coords


@_decorate_validation
def _validate_square_spiral_sample_image(h, w, scan_length, num_points):
    """
    Validate the `square_spiral_sample_image` function.

    See Also
    --------
    square_spiral_sample_image : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h, 'h', {'type': int, 'min': 2})
    _validate(w, 'w', {'type': int, 'min': 2})
    _validate(scan_length, 'scan_length', {'type': float,
                                           'min': _min_scan_length})
    _validate(num_points, 'num_points', {'type': int, 'min': _min_num_points})


def square_spiral_sample_image(h, w, scan_length, num_points):
    """
    Sample an image using a square spiral pattern.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a square spiral pattern are determined. The
    `scan_length` determines the length of the path scanned whereas
    `num_points` indicates the number of samples taken on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import square_spiral_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> square_spiral_sample_image(h, w, scan_length, num_points)
    array([[ 5.        ,  5.        ],
           [ 6.28571429,  5.97619048],
           [ 4.38095238,  3.71428571],
           [ 2.42857143,  5.92857143],
           [ 4.95238095,  7.57142857],
           [ 7.57142857,  6.02380952],
           [ 7.        ,  2.42857143],
           [ 2.83333333,  2.42857143],
           [ 1.14285714,  4.9047619 ],
           [ 1.35714286,  8.85714286],
           [ 5.52380952,  8.85714286],
           [ 8.85714286,  8.02380952]])

    """

    _validate_square_spiral_sample_image(h, w, scan_length, num_points)

    coords = square_spiral_sample_surface(float(h - 1), float(w - 1),
                                          scan_length, float(num_points), 1.0)
    coords = coords + 0.5

    return coords


@_decorate_validation
def _validate_square_spiral_sample_surface(l, w, speed, sample_rate, time):
    """
    Validate the `square_spiral_sample_surface` function.

    See Also
    --------
    square_spiral_sample_surface : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(l, 'l', {'type': float, 'min': _min_l})
    _validate(w, 'w', {'type': float, 'min': _min_w})
    _validate(speed, 'speed', {'type': float, 'min': _min_speed})
    _validate(sample_rate, 'sample_rate', {'type': float,
                                           'min': _min_sample_rate})
    _validate(time, 'time', {'type': float, 'min': _min_time})


def square_spiral_sample_surface(l, w, speed, sample_rate, time):
    """
    Sample a surface area using a square spiral pattern.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using a square spiral pattern are determined. The
    scanned path is determined from the probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import square_spiral_sample_surface
    >>> l = 1e-6
    >>> w = 1e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> square_spiral_sample_surface(l, w, speed, sample_rate, time)
    array([[  5.00000000e-07,   5.00000000e-07],
           [  4.00000000e-07,   4.00000000e-07],
           [  6.00000000e-07,   7.00000000e-07],
           [  5.00000000e-07,   3.00000000e-07],
           [  2.00000000e-07,   7.00000000e-07],
           [  8.00000000e-07,   7.00000000e-07],
           [  6.00000000e-07,   2.00000000e-07],
           [  1.00000000e-07,   4.00000000e-07],
           [  3.00000000e-07,   9.00000000e-07],
           [  9.00000000e-07,   8.00000000e-07],
           [  9.00000000e-07,   1.00000000e-07],
           [  2.00000000e-07,   1.00000000e-07]])

    """

    _validate_square_spiral_sample_surface(l, w, speed, sample_rate, time)

    samples = int(sample_rate * time)
    scan_length = speed * time
    sample_dist = scan_length / samples
    min_size = np.min([l, w])
    delta_dist = np.abs(l - w) / 2

    # Determine number of rotations and length of line segments in spiral
    roots = np.polynomial.polynomial.polyroots(
        [-(scan_length + 2 * delta_dist),
         min_size + 2 * delta_dist - scan_length,
         min_size + 2 * delta_dist])  # [c, b, a] in ax^2 + bx + c = 0
    N = np.ceil(roots.max())  # scan_length ~= (1+1+2+2+..+N+N) * line_length
    line_len = min_size / (N + 1)

    coords = np.zeros((samples, 2))
    position = np.zeros(2)  # Start spiral at (0, 0); later shift its position
    directions = np.array([[-1, 0], [0, 1], [1, 0], [0, -1]])  # -x, +y, +x, -y
    direction_num = 0  # Start spiral by a horisontal line to the left

    # Handle possibly non-square sampling area
    delta = delta_dist * np.array([w > l, w < l])

    n = 1  # n \in (range(N) + 1)
    k = 0  # k = 0 <-> horisontal line segment; k = 1 <-> vertical line segment
    for sample in range(samples):
        # Mark coordinates
        coords[sample, :] = position

        # Update position
        position = position + directions[direction_num, :] * sample_dist

        # Handle corners
        if np.abs(position[k]) > np.ceil(n / 2) * line_len + delta[k]:
            # Fix overshoot from last iteration
            wrap_dist = np.abs(position[k]) - (np.ceil(n / 2) * line_len +
                                               delta[k])
            position = position - directions[direction_num, :] * wrap_dist

            # Find offset if several corners are involved
            offset = np.zeros(2)
            while wrap_dist > n * line_len + 2 * delta[k - 1]:
                wrap_dist = wrap_dist - n * line_len + 2 * delta[k - 1]
                direction_num = np.mod(direction_num + 1, 4)
                k = np.mod(direction_num, 2)
                if k == 0:
                    n = n + 1
                offset = offset + ((n * line_len + 2 * delta[k]) *
                                   directions[direction_num, :])
            else:
                direction_num = np.mod(direction_num + 1, 4)
                k = np.mod(direction_num, 2)
                if k == 0:
                    n = n + 1

            position = position + offset
            position = position + directions[direction_num, :] * wrap_dist

    # Shift to center of region
    coords = coords + np.array([w / 2, l / 2])

    return coords


@_decorate_validation
def _validate_uniform_line_sample_image(h, w, scan_length, num_points):
    """
    Validate the `uniform_line_sample_image` function.

    See Also
    --------
    uniform_line_sample_image : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h, 'h', {'type': int, 'min': 2})
    _validate(w, 'w', {'type': int, 'min': 2})
    _validate(scan_length, 'scan_length', {'type': float,
                                           'min': _min_scan_length})
    _validate(num_points, 'num_points', {'type': int, 'min': _min_num_points})


def uniform_line_sample_image(h, w, scan_length, num_points):
    """
    Sample an image using a set of uniformly distributed straight lines.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a pattern based on a set of uniformly distributed
    straight lines are determined. The `scan_length` determines the length of
    the path scanned whereas `num_points` indicates the number of samples taken
    on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `scan_length` implies this. The top and bottom lines of the image are
    always included in the scan.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import uniform_line_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> uniform_line_sample_image(h, w, scan_length, num_points)
    array([[ 0.5       ,  0.5       ],
           [ 5.04545455,  0.5       ],
           [ 9.5       ,  0.59090909],
           [ 7.11363636,  2.75      ],
           [ 2.56818182,  2.75      ],
           [ 0.72727273,  5.        ],
           [ 5.27272727,  5.        ],
           [ 9.5       ,  5.31818182],
           [ 6.88636364,  7.25      ],
           [ 2.34090909,  7.25      ],
           [ 0.95454545,  9.5       ],
           [ 5.5       ,  9.5       ]])

    """

    _validate_uniform_line_sample_image(h, w, scan_length, num_points)

    coords = uniform_line_sample_surface(float(h - 1), float(w - 1),
                                         scan_length, float(num_points - 1),
                                         1.0)

    coords = coords + 0.5

    return coords


@_decorate_validation
def _validate_uniform_line_sample_surface(l, w, speed, sample_rate, time):
    """
    Validate the `uniform_line_sample_surface` function.

    See Also
    --------
    uniform_line_sample_surface : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(l, 'l', {'type': float, 'min': _min_l})
    _validate(w, 'w', {'type': float, 'min': _min_w})
    _validate(speed, 'speed', {'type': float, 'min': _min_speed})
    _validate(sample_rate, 'sample_rate', {'type': float,
                                           'min': _min_sample_rate})
    _validate(time, 'time', {'type': float, 'min': _min_time})


def uniform_line_sample_surface(l, w, speed, sample_rate, time):
    """
    Sample aa surface area using a set of uniformly distributed straight lines.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using a pattern based on a set of uniformly distributed
    straight lines are determined.  The scanned path is determined from the
    probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `l` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `scan_length` implies this. The top and bottom lines of the image are
    always included in the scan.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import uniform_line_sample_surface
    >>> l = 2e-6
    >>> w = 2e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> uniform_line_sample_surface(l, w, speed, sample_rate, time)
    array([[  0.00000000e+00,   0.00000000e+00],
           [  7.00000000e-07,   0.00000000e+00],
           [  1.40000000e-06,   0.00000000e+00],
           [  2.00000000e-06,   1.00000000e-07],
           [  1.86666667e-06,   6.66666667e-07],
           [  1.16666667e-06,   6.66666667e-07],
           [  4.66666667e-07,   6.66666667e-07],
           [  0.00000000e+00,   9.00000000e-07],
           [  2.66666667e-07,   1.33333333e-06],
           [  9.66666667e-07,   1.33333333e-06],
           [  1.66666667e-06,   1.33333333e-06],
           [  2.00000000e-06,   1.70000000e-06],
           [  1.60000000e-06,   2.00000000e-06]])

    """

    _validate_uniform_line_sample_surface(l, w, speed, sample_rate, time)

    samples = int(sample_rate * time) + 1
    scan_length = speed * time
    sample_dist = scan_length / (samples - 1)
    lines_scan_length = scan_length - l
    num_lines = np.ceil(lines_scan_length / w)

    # We should always at least partially scan top and bottom lines.
    if num_lines < 2:
        num_lines = 2

    lines = np.linspace(0, l, num=num_lines)

    coords = _get_line_scan_coords(lines, samples, sample_dist, l, w)

    return coords


@_decorate_validation
def _validate_unique_pixels(coords):
    """
    Validate the `unique_pixels` function.

    See Also
    --------
    unique_pixels : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate_ndarray(coords, 'coords', {'subdtype': np.floating, 'dim': 2})

    if coords.shape[1] != 2:
        raise TypeError('coords.shape[1] must be 2.')

    if coords.min() < 0:
        raise TypeError('All coordinates must be non-negative.')


def unique_pixels(coords):
    """
    Identify unique pixels from a set of coordinates.

    The floating point `coords` are reduced to a unique set of integer pixels
    by flooring the floating point values.

    Parameters
    ----------
    coords : ndarray
        The `k` floating point coordinates arranged into a 2D array where each
        row is a coordinate pair (x, y), such that `coords` has size `k` x 2.

    Returns
    -------
    unique_pixels : ndarray
        The `l` <= `k` unique (integer) pixels, such that `unique_pixels` is a
        2D array and has size `l` x 2.

    Examples
    --------
    For example,

    >>> from magni.imaging.measurements import unique_pixels
    >>> coords = np.array([[1.7, 1.0], [1.0, 1.2], [3.3, 4.3]])
    >>> unique_pixels(coords)
    array([[1, 1],
           [3, 4]])

    """

    _validate_unique_pixels(coords)

    pixels = np.floor(coords).astype(np.int64)

    pixels_struct = pixels.view(pixels.dtype.descr * 2)
    unique, index = np.unique(pixels_struct, return_index=True)
    unique_pixels = pixels[np.sort(index)]

    return unique_pixels


def _get_line_scan_coords(lines, samples, sample_dist, l, w):
    """
    Determine the coordinates of the sampled lines in a line scanning.

    Parameters
    ----------
    lines : ndarray
        The position (vertical distance from top of scan area) of the lines to
        scan.
    samples : int
        The number of samples on the scan path.
    sample_dist : float
        The distance (path length) between samples.
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.

    Returns
    -------
    coords : ndarray
        The w and l coordinates arranged into a N_samples-by-2 array.

    """

    dir_right = np.float64((1, 0))
    dir_left = np.float64((-1, 0))
    dir_up = np.float64((0, 1))

    coords = np.zeros((samples, 2))
    position = np.float64((0, 0))
    direction = dir_right

    dist_target = w
    dist_epsilon = 1e-6 * sample_dist
    line = 0

    for sample in range(1, samples):
        dist_remaining = sample_dist

        while dist_remaining > 0:
            if dist_remaining <= dist_target:
                position = position + dist_remaining * direction
                dist_target = dist_target - dist_remaining
                dist_remaining = 0
            elif np.abs(dist_remaining - dist_target) < dist_epsilon:
                position = position + dist_target * direction
                dist_target = 0
                dist_remaining = 0
            else:
                position = position + dist_target * direction
                # dist_target = 0
                dist_remaining = dist_remaining - dist_target

                if ((direction == dir_right).all() or
                        (direction == dir_left).all()):
                    direction = dir_up
                    dist_target = lines[line + 1] - lines[line]
                    line = line + 1
                else:
                    direction = dir_right if line % 2 == 0 else dir_left
                    dist_target = w
                    # line = line

        coords[sample, :] = position

    return coords



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing validation capability.

The intention is to validate all public functions of the package such that
erroneous arguments in calls are reported in an informative fashion rather than
causing arbitrary exceptions or unexpected results. To avoid performance
impairments, the validation can be disabled globally.

Routine listings
----------------
decorate_validation(func)
    Decorate a validation function (see Notes).
disable_validation()
    Disable validation globally (see Notes).
validate(var, path, levels, ignore_none=False)
    Validate the value of a variable according to a validation scheme.
validate_ndarray(var, path, constraints)
    Validate a numpy.ndarray according to a validation scheme.

Notes
-----
To be able to disable validation (and to ensure consistency), for every public
function a validation function with the same name prefixed by '_validate_'
should be defined. This function should be decorated by `decorate_validation`,
be placed just above the function which it decorates, and be called as the
first thing with all arguments.

Examples
--------
If, for example, the following function is defined:

>>> def greet(person, greeting):
...     print('{}, {} {}.'.format(greeting, person['title'], person['name']))

This function probably expects its argument, 'person' to be a dictionary with
keys 'title' and 'name' and its argument, 'greeting' to be a string. If, for
example, a list is passed as the first argument, a TypeError is raised with the
description 'list indices must be integers, not str'. While obviously correct,
this message is not excessively informative to the user of the function.
Instead, this module can be used to redefine the function as follows:

>>> from magni.utils.validation import validate, decorate_validation
>>> @decorate_validation
... def _validate_greet(person, greeting):
...     validate(person, 'person', {'type': dict, 'keys': ('title', 'name')})
...     validate(greeting, 'greeting', {'type': str})
>>> def greet(person, greeting):
...     _validate_greet(person, greeting)
...     print('{}, {} {}.'.format(greeting, person['title'], person['name']))

If, again, a list is passed as the first argument, a TypeError with the
description 'type(person) must be <type 'dict'>'. Now, the user of the function
can easily identify the mistake and correct the call to read:

>>> greet({'title': 'Mr.', 'name': 'Anderson'}, 'You look surprised to see me')
You look surprised to see me, Mr. Anderson.

"""

from __future__ import division

import numpy as np


_disabled = False


def decorate_validation(func):
    """
    Decorate a validation function (see module Notes).

    Parameters
    ----------
    func : function
        The validation function which should be decorated.

    Examples
    --------
    An example of a function which accepts only an integer as argument:

    >>> from magni.utils.validation import decorate_validation
    >>> @decorate_validation
    ... def _validate_test(arg):
    ...     magni.utils.validation.validate(arg, 'arg', {'type': int})
    >>> def test(arg):
    ...     _validate_test(arg)
    ...     return

    If the function is called with anything but an integer, it fails:

    >>> try:
    ...     test('string')
    ...     print('No exception occured')
    ... except BaseException:
    ...     print('An exception occured')
    An exception occured

    """

    def wrapper(*args, **kwargs):
        """
        Wrap a validation function (see module Notes).

        Parameters
        ----------
        args : tuple
            The arguments passed to the decorated function.
        kwargs : dict
            The keyword arguments passed to the decorated function.

        """

        if not _disabled:
            func(*args, **kwargs)

    return wrapper


def disable_validation():
    """
    Disable validation globally (see module Notes).

    There is no equivalent function to enable validation since either any or
    no function calls should be validated depending on the run mode.

    Examples
    --------
    An example of a function which accepts only an integer as argument:

    >>> @magni.utils.validation.decorate_validation
    ... def _validate_test(arg):
    ...     magni.utils.validation.validate(arg, 'arg', {'type': int})
    >>> def test(arg):
    ...     _validate_test(arg)
    ...     return

    If the function is called with anything but an integer, it fails:

    >>> try:
    ...     test('string')
    ...     print('No exception occured')
    ... except BaseException:
    ...     print('An exception occured')
    An exception occured

    However, if validation is disabled, the same call does not fail:

    >>> from magni.utils.validation import disable_validation
    >>> disable_validation()
    >>> try:
    ...     test('string')
    ...     print('No exception occured')
    ... except BaseException:
    ...     print('An exception occured')
    No exception occured

    """

    global _disabled
    _disabled = True


def validate(var, path, levels, ignore_none=False):
    """
    Validate the value of a variable according to a validation scheme.

    Parameters
    ----------
    var : None
        The variable to validate which can take any type.
    path : str
        The path of the variable which is printed if the variable is invalid.
    levels : list or tuple or dict
        The validation scheme.
    ignore_none : bool, optional
        The flag indicating whether the variable is allowed to have the value
        None (the default is False).

    Notes
    -----
    The `levels` parameter is either a dict or a sequence of dicts. If it is a
    dict, it is considered a sequence of one dict. Each dict corresponds to the
    validation scheme of a 'level': the first level consists of `var` itself,
    the second level consists of any value in `var` if `var` is itself a list,
    tuple, or dict, and so on.

    A level is validated according to the keys and values of the dict
    specifying the validation scheme of that level. See the source for details
    on the usable keys and their function.

    .. note:: It is assumed that `var`, `path`, `levels`, and `ignore_none`
              are all valid arguments as these are not explicitly validated.

    Examples
    --------
    Define a function for reporting if an exception occurs when calling making
    some call:

    >>> from magni.utils.validation import validate
    >>> def report(call):
    ...     try:
    ...         call()
    ...         print('No exception occured')
    ...     except BaseException:
    ...         print('An exception occured')

    An example of how to validate an integer:

    >>> var = 1
    >>> call = lambda: validate(var, 'var', {'type': int})
    >>> report(call)
    No exception occured

    The above code fails when, e.g., a string is passed instead:

    >>> var = 'string'
    >>> report(call)
    An exception occured

    An example of how to validate a tuple expected to hold two integers:

    >>> var = (42, 1337)
    >>> call = lambda: validate(var, 'var', ({'type': tuple, 'len': 2},
    ...                                      {'type': int}))
    >>> report(call)
    No exception occured

    The above code fails when, e.g., a tuple with three integers is passed
    instead:

    >>> var = (1, 2, 3)
    >>> report(call)
    An exception occured

    """

    if ignore_none and var is None:
        return

    if isinstance(levels, dict):
        levels = [levels]

    c = levels[0]

    if 'type' in c and not type(var) is c['type']:
        _error('type', 'type({!r}) must be {!r}.', (path, c['type']))

    if 'type_in' in c and not type(var) in c['type_in']:
        _error('type', 'type({!r}) must be in {!r}.', (path, c['type_in']))

    if 'class' in c and not isinstance(var, c['class']):
        _error('type', '{!r} must be an instance of {!r}.', (path, c['class']))

    if 'val_in' in c and var not in c['val_in']:
        _error('value', '{!r} must be in {!r}.', (path, c['val_in']))

    if 'min' in c and var < c['min']:
        _error('value', '{!r} must be >= {!r}.', (path, c['min']))

    if 'max' in c and var > c['max']:
        _error('value', '{!r} must be <= {!r}.', (path, c['max']))

    if 'len' in c and len(var) != c['len']:
        _error('value', '{!r} must be of length {!r}.', (path, c['len']))

    if 'keys' in c:
        for key in c['keys']:
            if key not in var:
                _error('key', '{!r} must have the key {!r}.', (path, key))

    if 'keys_in' in c:
        for key in var:
            if key not in c['keys_in']:
                _error('key', '{!r} may only have keys in {!r}.',
                       (path, c['keys_in']))

    if len(levels) > 1:
        if not isinstance(var, dict):
            for i, item in enumerate(var):
                validate(item, '{}[{}]'.format(path, i), levels[1:])
        else:
            for key, item in var.items():
                validate(item, '{}[{}]'.format(path, key), levels[1:])


def validate_ndarray(var, path, constraints={}, ignore_none=False):
    """
    Validate a numpy.ndarray according to a validation scheme.

    Parameters
    ----------
    var : numpy.ndarray
        The variable to validate.
    path : str
        The path of the variable which is printed if the variable is invalid.
    constraints : dict
        The validation scheme.
    ignore_none : bool, optional
        Boolean indicating whether the variable is allowed to have the value
        None (the default is False).

    Notes
    -----
    The variable is validated according to the keys and values of the dict
    specifying the validation scheme. See the source for details on the usable
    keys and their function.

    .. note:: It is assumed that `var`, `path`, `levels`, and `ignore_none`
              are all valid arguments as these are not explicitly validated.

    Examples
    --------
    Define a function for reporting if an exception occurs when calling making
    some call:

    >>> from magni.utils.validation import validate_ndarray
    >>> def report(call):
    ...     try:
    ...         call()
    ...         print('No exception occured')
    ...     except BaseException:
    ...         print('An exception occured')

    An example of how to validate a float64 numpy.ndarray of shape (5,):

    >>> var = np.float64([1, 2, 3, 4, 5])
    >>> call = lambda: validate_ndarray(var, 'var',
    ...                                 {'dtype': np.float64, 'shape': (5,)})
    >>> report(call)
    No exception occured

    The above code fails when, e.g., an int64 numpy.ndarray is passed instead:

    >>> var = np.int64([1, 2, 3, 4, 5])
    >>> report(call)
    An exception occured

    """

    if ignore_none and var is None:
        return

    c = constraints

    if not isinstance(var, np.ndarray):
        _error('type', '{!r} must be an instance of np.ndarray.', (path,))

    if 'dtype' in c and var.dtype.type != c['dtype']:
        _error('type', '{!r}.dtype must be {!r}.', (path, c['dtype']))

    if 'subdtype' in c and not np.issubdtype(var.dtype, c['subdtype']):
        _error('type', '{!r}.dtype must be a supdtype of {!r}.',
               (path, c['subdtype']))

    if 'shape' in c and var.shape != c['shape']:
        _error('type', '{!r}.shape must be {!r}.', (path, c['shape']))

    if 'dim' in c and len(var.shape) != c['dim']:
        _error('type', 'The dimension of {!r} must be {!r}.', (path, c['dim']))


def _error(error, message, args):
    """
    Raise an error of a given type with a given message.

    Parameters
    ----------
    error : {'type', 'value', 'key'}
        The type of the raised error.
    message : str
        The message of the raised error after being formatted using 'format'.
    args : list or tuple
        The args passed to 'format'.

    """

    if error == 'type':
        error = TypeError
    elif error == 'value':
        error = ValueError
    elif error == 'key':
        error = KeyError

    raise error(message.format(*args))



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing problem suite instance generation functionality.

The problem suite instances consist of a matrix, A, and a coefficient vector,
alpha, with which the measurement vector, y, can be generated.

Routine listings
----------------
generate_matrix(m, n)
    Generate a matrix belonging to a specific problem suite.
generate_vector(n, k)
    Generate a vector belonging to a specific problem suite.

Notes
-----
The matrices and vectors generated in this module use the numpy.random
submodule. Consequently, the calling script or function should control the seed
to ensure reproducibility.

Examples
--------
Generate a problem suite instance:

>>> from magni.cs.phase_transition import _data
>>> m, n, k = 400, 800, 100
>>> A = _data.generate_matrix(m, n)
>>> alpha = _data.generate_vector(n, k)
>>> y = np.dot(A, alpha)

"""

from __future__ import division

import numpy as np

from magni.cs.phase_transition import config as _config


def generate_matrix(m, n):
    """
    Generate a matrix belonging to a specific problem suite.

    The available ensemble is the Uniform Spherical Ensemble. See Notes for a
    description of the ensemble.

    Parameters
    ----------
    m : int
        The number of rows.
    n : int
        The number of columns.

    Returns
    -------
    A : ndarray
        The generated matrix.

    Notes
    -----
    The Uniform Spherical Ensemble:
        The matrices of this ensemble have i.i.d. Gaussian entries and its
        columns are normalised to have unit length.

    """

    A = np.float64(np.random.randn(m, n))

    for i in range(n):
        A[:, i] = A[:, i] / np.linalg.norm(A[:, i])

    return A


def generate_vector(n, k):
    """
    Generate a vector belonging to a specific problem suite.

    The available ensembles are the Gaussian ensemble and the Rademacher
    ensemble. See Notes for a description of the ensembles. Which of the
    available ensembles is used, is specified as a configuration option. Note,
    that the non-zero `k` non-zero coefficients are the `k` first entries.

    Parameters
    ----------
    n : int
        The length of the vector.
    k : int
        The number of non-zero coefficients.

    Returns
    -------
    alpha : ndarray
        The generated vector.

    See Also
    --------
    magni.cs.phase_transition.config : Configuration options.

    Notes
    -----
    The Gaussian ensemble :
        The non-zero coefficients are drawn from the normal Gaussian
        distribution.
    The Rademacher ensemble:
        The non-zero coefficients are drawn from the constant amplitude with
        random signs ensemble.

    """

    x = np.zeros((n, 1))
    coefficients = _config.get('coefficients')

    if coefficients == 'rademacher':
        x[:k, 0] = np.random.randint(0, 2, k) * 2 - 1
    elif coefficients == 'gaussian':
        x[:k, 0] = np.random.randn(k)

    return x



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing matrix emulators.

The matrix emulators of this module are wrappers of fast linear operations
giving the fast linear operations the same basic interface as a numpy ndarray.
Thereby allowing fast linear operations and numpy ndarrays to be used
interchangably in other parts of the package.

Routine listings
----------------
Matrix()
    Wrap fast linear operations in a matrix emulator.
MatrixCollection()
    Wrap multiple matrix emulators in a single matrix emulator.

See Also
--------
magni.imaging._fastops : Fast linear operations.

"""

from __future__ import division
import types

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate
from magni.utils.validation import validate_ndarray as _validate_ndarray


class Matrix():
    """
    Wrap fast linear operations in a matrix emulator.

    `Matrix` defines a few attributes and internal methods which ensures that
    instances have the same basic interface as a numpy matrix instance without
    explicitly forming the matrix. This basic interface allows instances to be
    multiplied with vectors, to be transposed, and to assume a shape. Also,
    instances have an attribute which explicitly forms the matrix.

    Parameters
    ----------
    func : function
        The fast linear operation applied to the vector when multiplying the
        matrix with a vector.
    trans : function
        The fast linear operation applied to the vector when multiplying the
        transposed matrix with a vector.
    args : list or tuple
        The arguments which should be passed to `func` and `trans` in addition
        to the vector.
    shape : list or tuple
        The shape of the emulated matrix.

    Examples
    --------
    For example, the negative identity matrix could be emulated as

    >>> from magni.utils.matrices import Matrix
    >>> func = lambda vec: -vec
    >>> matrix = Matrix(func, func, (), (3, 3))

    The example matrix will have the desired shape:

    >>> matrix.shape
    (3, 3)

    The example matrix will behave just like an explicit matrix:

    >>> vec = np.float64([1, 2, 3]).reshape(3, 1)
    >>> matrix.dot(vec)
    array([[-1.],
           [-2.],
           [-3.]])

    If, at some point, an explicit representation of the matrix is required,
    this can easily be obtained:

    >>> matrix.A
    array([[-1., -0., -0.],
           [-0., -1., -0.],
           [-0., -0., -1.]])

    Likewise, the transpose of the matrix can be obtained:

    >>> matrix.T.A
    array([[-1., -0., -0.],
           [-0., -1., -0.],
           [-0., -0., -1.]])

    """

    @_decorate_validation
    def _validate_init(self, func, trans, args, shape):
        """
        Validate the `__init__` function.

        See Also
        --------
        Matrix.__init__ : The validated function.
        magni.utils.validation.validate : Validation.

        """

        _validate(func, 'func', {'type': types.FunctionType})
        _validate(trans, 'trans', {'type': types.FunctionType})
        _validate(args, 'args', {'type_in': (list, tuple)})
        _validate(shape, 'shape', [{'type_in': (list, tuple)}, {'type': int}])

    def __init__(self, func, trans, args, shape):
        self._validate_init(func, trans, args, shape)

        self._func = func
        self._trans = trans
        self._args = args
        self._shape = shape

    @property
    def A(self):
        """
        Explicitly form the matrix.

        The fast linear operations implicitly define a matrix which is usually
        not explicitly formed. However, some functionality might require a more
        advanced matrix interface than that provided by this class.

        Returns
        -------
        matrix : numpy.ndarray
            The explicit matrix.

        Notes
        -----
        The explicit matrix is formed by multiplying the matrix with the
        columns of an identity matrix and stacking the resulting vectors as
        columns in a matrix.

        """

        output = np.zeros(self.shape, dtype=np.complex128)
        vec = np.zeros((self.shape[1], 1))

        for i in range(self.shape[1]):
            vec[i] = 1
            output[:, i] = self.dot(vec)[:, 0]
            vec[i] = 0

        if np.abs(output.imag).sum() == 0:
            output = output.real

        return output

    @property
    def shape(self):
        """
        Get the shape of the matrix.

        Returns
        -------
        shape : tuple
            The shape of the matrix.

        """

        return self._shape

    @property
    def T(self):
        """
        Get the transpose of the matrix.

        Returns
        -------
        matrix : Matrix
            The transpose of the matrix.

        Notes
        -----
        The fast linear operation and the fast linear transposed operation of
        the resulting matrix are same as those of the current matrix except
        swapped. The shape is modified accordingly.

        """

        return Matrix(self._trans, self._func, self._args, self._shape[::-1])

    @_decorate_validation
    def _validate_dot(self, vec):
        """
        Validate the `dot` function.

        See Also
        --------
        Matrix.dot : The validated function.
        magni.utils.validation.validate_matrix : Validation.

        """

        _validate_ndarray(vec, 'vec', {'shape': (self.shape[1], 1)})

    def dot(self, vec):
        """
        Multiply the matrix with a vector.

        Parameters
        ----------
        vec : numpy.ndarray
            The vector which the matrix is multiplied with.

        Returns
        -------
        vec : numpy.matrix
            The result of the multiplication.

        """

        self._validate_dot(vec)

        return self._func(vec, *self._args)


class MatrixCollection():
    """
    Wrap multiple matrix emulators in a single matrix emulator.

    `MatrixCollection` defines a few attributes and internal methods which
    ensures that instances have the same basic interface as a numpy matrix
    instance without explicitly forming the matrix. This basic interface allows
    instances to be multiplied with vectors, to be transposed, and to assume a
    shape. Also, instances have an attribute which explicitly forms the matrix.

    Parameters
    ----------
    matrices : list or tuple
        The collection of `Matrix` instances.

    See Also
    --------
    Matrix : Matrix emulator.

    Examples
    --------
    For example, two matrix emulators can be combined into one. That is, the
    matrix:

    >>> func = lambda vec: -vec
    >>> negate = magni.utils.matrices.Matrix(func, func, (), (3, 3))
    >>> negate.A
    array([[-1., -0., -0.],
           [-0., -1., -0.],
           [-0., -0., -1.]])

    And the matrix:

    >>> func = lambda vec: vec[::-1]
    >>> reverse = magni.utils.matrices.Matrix(func, func, (), (3, 3))
    >>> reverse.A
    array([[ 0.,  0.,  1.],
           [ 0.,  1.,  0.],
           [ 1.,  0.,  0.]])

    Can be combined into one matrix emulator using the present class:

    >>> from magni.utils.matrices import MatrixCollection
    >>> matrix = MatrixCollection((negate, reverse))

    The example matrix will have the desired shape:

    >>> matrix.shape
    (3, 3)

    The example matrix will behave just like an explicit matrix:

    >>> vec = np.float64([1, 2, 3]).reshape(3, 1)
    >>> matrix.dot(vec)
    array([[-3.],
           [-2.],
           [-1.]])

    If, at some point, an explicit representation of the matrix is required,
    this can easily be obtained:

    >>> matrix.A
    array([[-0., -0., -1.],
           [-0., -1., -0.],
           [-1., -0., -0.]])

    Likewise, the transpose of the matrix can be obtained:

    >>> matrix.T.A
    array([[-0., -0., -1.],
           [-0., -1., -0.],
           [-1., -0., -0.]])

    """

    @_decorate_validation
    def _validate_init(self, matrices):
        """
        Validate the `__init__` function.

        See Also
        --------
        MatrixCollection.__init__ : The validated function.
        magni.utils.validation.validate : Validation.

        """

        _validate(matrices, 'matrices',
                  [{'type_in': (list, tuple)}, {'class': Matrix}])

        for i in range(len(matrices) - 1):
            if matrices[0].shape[1] != matrices[1].shape[0]:
                raise ValueError('The matrices must have compatible shapes.')

    def __init__(self, matrices):
        self._validate_init(matrices)

        self._matrices = matrices

    @property
    def A(self):
        """
        Explicitly form the matrix.

        The collection of matrices implicitly defines a matrix which is usually
        not explicitly formed. However, some functionality might require a more
        advanced matrix interface than that provided by this class.

        Returns
        -------
        matrix : numpy.ndarray
            The explicit matrix.

        Notes
        -----
        The explicit matrix is formed by multiplying the matrix with the
        columns of an identity matrix and stacking the resulting vectors as
        columns in a matrix.

        """

        output = np.zeros(self.shape, dtype=np.complex128)
        vec = np.zeros((self.shape[1], 1))

        for i in range(self.shape[1]):
            vec[i] = 1
            output[:, i] = self.dot(vec)[:, 0]
            vec[i] = 0

        if np.abs(output.imag).sum() == 0:
            output = output.real

        return output

    @property
    def shape(self):
        """
        Get the shape of the matrix.

        Returns
        -------
        shape : tuple
            The shape of the matrix.

        Notes
        -----
        The shape of the product of a number of matrices is the number of rows
        of the first matrix times the number of columns of the last matrix.

        """

        return (self._matrices[0].shape[0], self._matrices[-1].shape[1])

    @property
    def T(self):
        """
        Get the transpose of the matrix.

        Returns
        -------
        matrix : MatrixCollection
            The transpose of the matrix.

        Notes
        -----
        The transpose of the product of the number of matrices is the product
        of the transpose of the matrices in reverse order.

        """

        return MatrixCollection([matrix.T for matrix in self._matrices[::-1]])

    @_decorate_validation
    def _validate_dot(self, vec):
        """
        Validate the `dot` function.

        See Also
        --------
        MatrixCollection.dot : The validated function.
        magni.utils.validation.validate_matrix : Validation.

        """

        _validate_ndarray(vec, 'vec', {'shape': (self.shape[1], 1)})

    def dot(self, vec):
        """
        Multiply the matrix with a vector.

        Parameters
        ----------
        vec : numpy.matrix
            The vector which the matrix is multiplied with.

        Returns
        -------
        vec : numpy.matrix
            The result of the multiplication.

        """

        self._validate_dot(vec)

        for matrix in self._matrices[::-1]:
            vec = matrix.dot(vec)

        return vec



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public function of the magni.cs.phase_transition
subpackage.

"""

from __future__ import division
import os
import re
import types

from magni.cs.phase_transition import _analysis
from magni.cs.phase_transition import _simulation
from magni.utils.multiprocessing import File as _File
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate


@_decorate_validation
def _validate_determine(algorithm, path, label, overwrite):
    """
    Validate the `determine` function.

    See Also
    --------
    determine : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(algorithm, 'algorithm', {'type': types.FunctionType})
    _validate(path, 'path', {'type': str})
    _validate(label, 'label', {'type': str})

    # regular expression matching invalid characters
    regexp = re.compile(r'[^a-zA-Z0-9 ,.\-_/]')

    if regexp.search(label):
        raise RuntimeError("label may not contain {!r}."
                           .format(regexp.search(label).group()))

    # regular expression matching labels without empty path components
    regexp = re.compile(r'^([^/]+/)*[^/]+$')

    if not regexp.search(label):
        raise RuntimeError('label may not contain empty path components.')

    _validate(overwrite, 'overwrite', {'type': bool})


def determine(algorithm, path, label='default', overwrite=False):
    """
    Determine the phase transition of a reconstruction algorithm.

    The phase transition is determined from a number of monte carlo simulations
    on a delta-rho grid for a given problem suite.

    Parameters
    ----------
    algorithm : function
        A function handle to the reconstruction algorithm.
    path : str
        The path of the HDF5 database where the results should be stored.
    label : str
        The label assigned to the phase transition (the default is 'default').
    overwrite : bool
        A flag indicating if an existing phase transition should be overwritten
        if it has the same path and label (the default is False).

    See Also
    --------
    magni.cs.phase_transition.config : Configuration options.
    magni.cs.phase_transition._simulation.run : The actual simulation.
    magni.cs.phase_transition._analysis.run : The actual phase determination.

    Examples
    --------
    An example of how to use this function is provided in the `examples` folder
    in the `cs-phase_transition.ipynb` ipython notebook file.

    """

    _validate_determine(algorithm, path, label, overwrite)

    if os.path.isfile(path):
        with _File(path, 'r') as f:
            if '/' + label in f:
                if overwrite:
                    f.remove_node('/' + label, recursive=True)
                else:
                    raise IOError("{!r} already uses the label {!r}."
                                  .format(path, label))

    _simulation.run(algorithm, path, label)
    _analysis.run(path, label)
